function bernstein_polynomial(x, n, N)
  return binomial(N, n) / 2^N * (1 + x)^n * (1 - x)^(N-n)
end


function vandermonde_matrix_bernstein(z)
  Npts = length(z)
  V = [ bernstein_polynomial(z[i], j-1, Npts-1) for i=1:Npts, j=1:Npts ]
  return V
end


struct BernsteinReconstruction
  P::Matrix # projection matrix to transfer data onto Bernstein nodes
  V::Matrix # vandermonde matrix to resum Bernstein series on nodes

  function BernsteinReconstruction(P, I)
    # reconstruction should maintain grid structure
    @assert size(P)[1] == size(I)[2]
    @assert size(P)[2] == size(I)[1]
    return new(P,I)
  end
end


function BernsteinReconstruction(x)
  Npts = length(x)
  rng = collect(0:Npts-1)
  bernstein_x = @. -1 + 2 * rng / (Npts-1)
  P = dg1d.lagrange_interp_matrix(bernstein_x, x)
  V = dg1d.vandermonde_matrix_bernstein(x)
  return BernsteinReconstruction(P, V)
end


@doc """
  reconstruct!(u!, w, bernstein::BernsteinReconstruction; kwargs...)

Compute inplace a convex combination of `u!` and its Bernstein reconstruction 
weighted by `w`.  
`m` and `M` keywords allow to enforce lower/upper bound in reconstruction.
"""
function reconstruct!(u!, w, bernstein::BernsteinReconstruction;
  m=missing, M=missing)

  @unpack P, V = bernstein
  uB = P * u!

  if !ismissing(m)
    @. uB = max(uB, m)
  end
  if !ismissing(M)
    @. uB = min(uB, M)
  end

  u! .*= w
  u! .+= (1-w) * V * uB
  return
end
