module FD



@doc """
Second order centered finite difference approximation for
  \frac{df}{dx}
on unstructured grid. 
Implementation is based on a second order Lagrange inteprolating polynomial
that connects data points around each grid point.
Boundary points are treated with one-sided differences of the same order.

f ... data points
x ... grid points
"""
function central_dx(f, x)

  N = length(f)
  if N <= 1
    error("Require at least two data points!")
  end

  df = zeros(N)
  dx = diff(x)

  for i = 2:N-1
    dxm    = dx[i-1]
    dxp    = dx[i]
    sumdx  = dxm + dxp
    df[i]  += - dxp / (dxm * sumdx) * f[i-1]
    df[i]  += (dxp - dxm) / (dxp * dxm) * f[i]
    df[i]  += dxm / (dxp * sumdx) * f[i+1]
  end

  # forward/left-directed difference
  dxm = dx[1]
  dxp = dx[2]
  sumdx = dxm + dxp
  df[1] += - (2 * dxm + dxp) / (sumdx * dxm) * f[1]
  df[1] += sumdx / (dxm * dxp) * f[2]
  df[1] += - dxm / (dxp * sumdx) * f[3]

  # forward/left-directed difference
  dxm = dx[end-1]
  dxp = dx[end]
  sumdx = dxm + dxp
  df[N] += dxp / (dxm * sumdx) * f[N-2]
  df[N] += - sumdx / (dxp * dxm) * f[N-1]
  df[N] += (2 * dxp + dxm) / (sumdx * dxp) * f[N]

  return df
end



end # FD
