# Design considerations
#
#   - I want to reuse as many parameter structs as possible, because
#   of DRY. And this also simplifies/unifies documentation.
#
#   - Problem is that some modules will require different parameters (or
#   only a subset) to work, e.g. AV module and all version thereof.
#   Q: Would be good to make a parameter struct for each AV method separately?
#     +: Make it clear which parameters to be used for each method.
#     -: Risk of DRY, because e.g. MDA uses a subset of parameters of MDH.
#
#   - IO: Use TOML and dictionaries as intermediate communication format.
#     +: Julia has native TOML support which parses automatically to dictionaries.
#     -: ...
#     Q: What were the reason to prefer TOML over YAML?
#
#   - DSL: TOML headers and parameter families?
#   Ideally, I want to establish unique TOML headers which will then map
#   into parameter structs.
#   Q: How to deal with nested parameter structs?
#   Q: How to document a given module parameter sturct is composed of 
#   sub-parameter structs?
#
#   Naive idea: Use Module name as TOML header(actually called tables) 
#   which then maps into
#   a corresponding parameter struct which is then used for all functionalities
#   of a module.
#   Q: What to do if multiple instances of a parameter struct are needed?
#   E.g. MDA1, MDA2, ...?
#   PA: Build parameter structs with meaninungful and non-duplicated 
#   member names (this is ensured by the language) and use those names as
#   TOML headers. TOML parsing happens separated from constructing the 
#   parameter struct. So the user then has to explicitly specify which parameter
#   struct he wants to construct out of the resulting dictionary.
#   The parameter struct must be fully type annotated so that one can figure
#   out dynamically which sub-parameter struct must be called if a certain
#   key is found in a dictionary.
#     -: This might introduce inconsistency between projects, e.g. one
#     names the evolution struct evolve, the other calls it time_stepping ...
#     Q: Is this really a problem if one documents the member names properly?
#     PA: Separate projects are separate, should one really care about what
#     others do? 
#     If I would not want this, then I should use the module
#     name, because that will be unqiue. Problem with module names is that the
#     scope of nested modules will have to be accounted for in the header
#     naming. Provided we only use the actual module name, without module
#     scope, then this could break uniqueness, because different modules
#     could contain submodules that are named the same.
#     Also: What if multiple parameter modules are needed in a module? This
#     would not work with using the module name as a unique identifier.
#
#   - Q: How to do parameter validation?
#     A: This sould happen in the sub-parameter constructors.
#   
#   - Q: Could this become a separate Julia package?
#     Q: How much does this really differ from Configurations.jl?


export save, load, AbstractParameters, sanity_check


@doc """
Supertype for Parameter structs to aid in converting their members
into Dictionaries compatible with the TOML format.

Example:
```julia
import Parameters: @with_kw

@with_kw struct Parameters_TimeStepping <: AbstractParameters
  tstart = 0.0
  tend = 1.0
  nsteps = 100

  @assert tstart < tend
  @assert iseven(nsteps)
end
```
"""
abstract type AbstractParameters end

# forward decleration
function to_dict end

# Helper to convert `key, value` pair to dictionary element that is compatible with 
# TOML format. Allowed TOML types are
#   - Number
#   - String
# and `Vector`s thereof.
# We add a version for `AbstractParameters` that calls the `to_dict` method
# that in turn calls `to_dict_element`. Eventually, all types must be
# compatible TOML types.
function to_dict_element(s::String, x::Union{T,Vector{T}}) where {T<:Union{Number,String}} 
  return Pair(s, x)
end
function to_dict_element(s::String, x::AbstractParameters)
  return Pair(s, to_dict(x))
end


# Convert a parameter struct to a TOML conforming dictionary of type
# Dict{String,Any}.
function to_dict(prms::AbstractParameters)
  Dict(to_dict_element(string(fn), getfield(prms, fn)) for fn in fieldnames(typeof(prms)))
end


@doc """
  `save(prms::AbstractParameters, filename)`

Save the fields of a `AbstractParameters` struct to a TOML file `filename`.
"""
function save(prms::AbstractParameters, filename)
  dict_prms = to_dict(prms)
  open(filename, "w") do io
    TOML.print(io, dict_prms)
  end
  return
end


# Convert any `Dict` into a `Dict{Symbol,Any}` dictionary. Used for converting
# a `Dict{String,Any}`, which is the result of TOML.parsefile, so that
# it can be handed to the keyword constructor generated by the 
# Parameters.@with_kw macro.
function to_symbolic_dict(dict::Dict)
  symbolic_dict = Dict{Symbol,Any}()
  for (key,value) in dict
    if (typeof(value) <: Dict)
      symbolic_dict[Symbol(key)] = to_symbolic_dict(value)
    else
      symbolic_dict[Symbol(key)] = value
    end
  end
  return symbolic_dict
end


@doc """
  `load(::Type{<:AbstractParameters}, filename)`

Load a TOML file `filename` into a parameter struct.
"""
function load(::Type{T}, filename) where T<:AbstractParameters
  dict_prms = TOML.parsefile(filename)
  symbolic_dict_prms = to_symbolic_dict(dict_prms)
  return from_dict(T, symbolic_dict_prms)
end


function from_dict(::Type{T}, args::Dict{Symbol, Any}) where T<:AbstractParameters
  typed_args = deepcopy(args)
  for (fn, ft) in zip(fieldnames(T), fieldtypes(T))
    if ft <: AbstractParameters
      typed_args[fn] = from_dict(ft, args[fn])
    elseif ft == Any
      continue
    else
      typed_args[fn] = ft(args[fn])
    end
  end
  return T(; typed_args...)
end


# # Taken from Parameters.jl
# # If @with_kw_noshow is used, then
# function Base.show(io::IO, p::T) where T <: AbstractParameters
#   if get(io, :compact, false) || get(io, :typeinfo, nothing)==T
#     Base.show_default(IOContext(io, :limit => true), p)
#   else
#     # just dumping seems to give ok output, in particular for big data-sets:
#     dump(IOContext(io, :limit => true), p)
#   end
# end

#######################################################################
#                                TODOS                                #
#######################################################################

#
# function Base.isequal(prms1::T, prms2::T) where T<:AbstractParameters
#   trues = Vector{Bool}(undef, 0)
#   for fn in fieldnames(prms1)
#     display(fn)
#     push!(trues, getfield(prms1, fn) == getfield(prms2, fn))
#   end
#   return all(trues)
# end
#
# # Write logic to deal with lists of string elements within a single string.
# # e.g. "var1 var2 var3" should be interpreted as ["var1", "var2", "var3"].
# # To do this in TOML one would have to quote every string separately.
