import Plots: mm


export @debug_plot


function grid_layout(nplts)
  if nplts == 1
    return (1,1)
  elseif iseven(nplts)
    return (Int64(nplts/2), 2)
  else
    return (Int64((nplts-1)/2), 3)
  end
end


function plot(vars::AbstractVariables, mesh::Mesh;
    fn=(u)->u, marker=(:circle), title="", 
    kwargs...)

  @unpack x, K = mesh
  @unpack Npts = mesh.element

  xlabel = "r"
  ylabels = [ "$(string(f))(t, r)" for f in fieldnames(typeof(vars)) ]
  vec_vars = fields(vars)

  nplts = length(vec_vars)
  plts = [ Plots.plot() for _ in 1:nplts ]

  for (plt, var, ylabel) in zip(plts, vec_vars, ylabels)
    mat_var = reshape(var, 1:Npts, 1:K)
    for k in 1:K
      Plots.plot!(plt, x[:,k], 
                  fn.(mat_var[:,k]);
                  label=false,
                  marker=marker,
                  kwargs...)
    end
    Plots.plot!(plt, ylabel=ylabel, xlabel=xlabel)
  end

  dict_kwargs = Dict(kwargs...)
  layout = get(dict_kwargs, :layout, grid_layout(nplts))
  size = pop!(dict_kwargs, :size, (1400,1000))

  plt = Plots.plot(plts..., layout=layout, plot_title=title, 
                   size=size, left_margin=10mm)
  return plt
end



function plot(segment::Segment{T}, mesh::Mesh;
    fn=(u)->u, marker=(:circle), title="", 
    kwargs...) where T<:AbstractVariables

  @unpack mesh, pts, cells = segment
  @unpack x, K = mesh
  @unpack Npts = mesh.element

  sub_x = x[pts, cells]

  xlabel = "r"
  ylabels = [ "$(string(f))(t, r)" for f in fieldnames(typeof(segment.var)) ]
  vec_vars = fields(segment.var)

  nplts = length(vec_vars)
  plts = [ Plots.plot() for _ in 1:nplts ]


  for (plt, var, ylabel) in zip(plts, vec_vars, ylabels)
    mat_var = reshape(var, 1:Npts, 1:K)
    sub_var = mat_var[pts, cells]
    for k in 1:(cells.stop-cells.start+1)
      Plots.plot!(plt, sub_x[:,k], 
                  fn.(sub_var[:,k]);
                  label=false,
                  marker=marker,
                  kwargs...)
    end
    Plots.plot!(plt, ylabel=ylabel, xlabel=xlabel)
  end

  dict_kwargs = Dict(kwargs...)
  layout = get(dict_kwargs, :layout, grid_layout(nplts))
  size = pop!(dict_kwargs, :size, (1400,1000))

  plt = Plots.plot(plts..., layout=layout, plot_title=title, 
                   size=size, left_margin=10mm)

  return plt
end


# overload with AbstractArray
function plot_dg(u, mesh::Mesh; kwargs...)

  @unpack x, K = mesh
  @unpack Npts = mesh.element

  plt = Plots.plot()

  mat_u = reshape(u, 1:Npts, 1:K)
  for k in 1:K
    Plots.plot!(plt, x[:,k], mat_u[:,k]; 
                label=false, kwargs...)
  end

  return plt
end


# overload with Vector of Interfaces
function plot_interfaces(vec_lhs, vec_rhs, mesh::Mesh;
    kwargs...)
  @unpack x, K = mesh
  @unpack Npts = mesh.element

  @assert length(vec_lhs) == length(vec_rhs)

  nplts = length(vec_lhs)

  Plots.gr()
  plts = [ Plots.plot() for _ in 1:nplts ]

  xlhs = mesh.x[1,:]
  xrhs = mesh.x[end,:]

  vars = [ vec_lhs, vec_rhs ]
  xs = [ xlhs, xrhs]

  xlabel = "r"
  ylabels = [ "lhs", "rhs" ]
  markers = [ :circle, :star ]

  for (i, plt) in enumerate(plts)
    for (var, x, ylabel, marker) in zip(vars, xs, ylabels, markers)
      scatter!(plt, x, var[i];
               label=ylabel, marker = (marker),
               kwargs...)
    end
  end

  dict_kwargs = Dict(kwargs)
  layout = pop!(dict_kwargs, :layout, grid_layout(nplts))
  size = pop!(dict_kwargs, :size, (1400,1000))
  plt = Plots.plot(plts..., layout=layout, size=size)

  return plt
end



macro debug_plot(vars, mesh, kwargs=())
  str_vars = string(vars)
  esc(
    quote
      plt = dg1d.plot($vars, $mesh; title=$str_vars*"::$(typeof($vars))", $kwargs... )
      Plots.gui(plt)
      readline()
    end
  )
end


#### Draft for plotting with Makie, but this is currently too slow ...
#
# import GLMakie
#
# function makie_plot(vars::AbstractVariables, mesh::Mesh;
#     fn=(u)->u, layout=nothing, kwargs...)
#
#   @unpack x, K = mesh
#   @unpack Npts = mesh.element
#
#   xlabel = "r"
#   ylabels = [ "$(string(f))(t, r)" for f in fieldnames(typeof(vars)) ]
#   vec_vars = fields(vars)
#
#   nplts = length(vec_vars)
#   # isnothing(layout) && layout = (nplts,)
#
#   fig = GLMakie.Figure()
#
#   # for (plt, var, ylabel) in zip(plts, vec_vars, ylabels)
#   for (idx, var) in enumerate(vec_vars)
#     mat_var = reshape(var, 1:Npts, 1:K)
#     ax = GLMakie.Axis(fig[idx,1])
#     for k in 1:K
#       clr = cell_colors[periodic_index(k, length(cell_colors))]
#       GLMakie.scatterlines!(ax, x[:,k], fn.(mat_var[:,k]), color=clr)
#     end
#     # Plots.plot!(plt, ylabel=ylabel, xlabel=xlabel)
#   end
#
#   GLMakie.current_figure()
#
#   # plt = Plots.plot(plts..., layout=(nplts,1))
#   # display(plt)
#
#   return
# end


