export AV
module AV


using dg1d, LinearAlgebra
import dg1d: Abstract_Eq, Mesh
import Parameters: @unpack, @with_kw


@doc """
Interface to Artificial viscosity method.

Artificial viscosity information that aims to resolve shocks and discontinuities
by regularizing a first order hyperbolic conservation law with a second
order spatial derivative term, e.g.
  
  ∂t u + ∂x f(u) = 0   =>   ∂t u + ∂x f(u) = ∂x(ν(x) ∂x u)

# TODO Think this through again.
AV methods must implement the following routines
AV(method, prms) ... constructor for AV method
compute!(AV, state, pblm) ... compute AV for current state.
"""
abstract type Abstract_AV end


export compute
@doc """
_TODO_ Reorder arguments in interface to match docstring.

  compute(av::Abstract_AV, u, v, t, mesh::Mesh)

Compute artificial viscosity from model `av` inplace.

# Arguments
- `av::Abstract_AV`: an artificial viscosity model.
- `u`: scalar field.
- `v`: advection speed, f'(u).
- `t`: current time.
"""
function compute end


struct DummyEq <: Abstract_Eq end
function compute(u, v, t, mesh::Mesh, av::Abstract_AV)
  return compute(u, v, t, mesh, DummyEq(), av)
end


@with_kw struct Parameters
  method::Symbol          = :MDA
  cmax::Real              = 1.0
  tstart::Real            = 0.0
  c1::Union{Real,Nothing} = nothing
  c2::Union{Real,Nothing} = nothing
end


include("mda.jl")
include("mdh.jl")
include("db.jl")
include("ev.jl")


@doc """
Construct Artifical Viscosity based on parameters and current mesh.

FIXME This is type instable.
Maybe use a parameteric struct approach for AV with Val as dispatch type.
"""
function make_AV(prms::Parameters, mesh::Mesh)
  ## need proper error handling in case eval(prms.method) is not defined
  return eval(prms.method)(prms, mesh)
end


@doc """
TODO Agree on one time step computation.
TODO Implement adaptive time stepping computation.
"""
function timestep(u, abs_v, mesh::Mesh, av::Abstract_AV)
  @unpack element, CFL, dl, x = mesh
  @unpack N, z = element
  @unpack cmax = av

  vmax = maximum(abs_v)
  μmax = maximum(av.μ[:])
  dt = CFL * dl/(N^2 * vmax)

  # Time step size of (2.1) in Kloeckner 2011
  # gives very very small time steps
  # TODO cmax shoudl actually be μ_max, Q: how to compute this one?
  # dt = CFL / (vmax * N^2 / dl + μmax * N^4 / dl^2)

  return dt
end


@doc """
TODO Replace with above method, e.g. take `abs_v` as separate input.
This should allow for better modularization.
"""
function timestep(u, mesh::Mesh, eq::Abstract_Eq, av::Abstract_AV)
  @unpack x = mesh
  abs_v = dg1d.maxabs_v(eq, u, x)
  return timestep(u, abs_v, mesh, av)
end


end
