@doc """
Entropy Viscosity model, cf. Yu, Hesthaven, 2019.
"""
@with_kw mutable struct EV <: Abstract_AV @deftype Real
  cmax
  ce
  t_prev
  u_prev::AbstractArray
  μ::AbstractArray
end


@doc """
Construct `EV` object based on mesh.
"""
function EV(prms::Parameters, mesh::Mesh)
  @unpack x = mesh
  u0 = zeros(Float64, size(x))
  μ0 = zeros(Float64, size(x))
  return EV(prms.cmax, prms.c1, prms.tstart, u0, μ0)
end


function compute(u, v, t, mesh::Mesh, eq::Abstract_Eq, av::EV)
  @unpack element, K, L, x, invjac = mesh
  @unpack D, N = element
  @unpack u_prev, t_prev, cmax, ce = av

  E       = @. Eq.E(u, x)
  E_prev  = @. Eq.E(u_prev, x)
  F       = @. Eq.F(u, x)
  F_prev  = @. Eq.F(u_prev, x)
  Δt      = t - t_prev

  # h ... 'characteristic' length of sub cells (Yu, Hesthaven, 2019)
  # consistent with Zingan et al., 2013
  h = L / (K * N)

  # entropy residual violation
  # Q: Is it really justified to use no viscosity if we have to compute
  # entropy wrt backwards time step?
  if Δt > 0.0
    residual = (E - E_prev) / Δt + 0.5 * D * (F + F_prev) .* invjac
  else
    residual = zeros(size(u))
  end
  R = vec(maximum(abs.(residual), dims=1))

  # entropy jump at interfaces
  intf_F = dg1d.to_interface(F, element, mesh)
  # TODO What about bdry conditions?
  jump = abs.(intf_F.lhs - intf_F.rhs) * N / h
  # extract jumps occuring at lhs and rhs of a cell's interface
  jump_lhs = jump[2:end]
  jump_rhs = jump[1:end-1]

  avg_E = dg1d.grid_average(E, element, mesh)
  norm_E = maximum(@. abs(E - avg_E))

  max_R = max.(R, jump_lhs, jump_rhs)
  μ_e = max_R / norm_E * ce * (h / N)^2

  # upper bound on AV
  maxabs_v = vec(maximum(abs.(v), dims=1))
  μ_max = @. cmax * maxabs_v * h / N

  # final AV
  μ = min.(μ_e, μ_max)

  # update history for u
  av.u_prev .= u
  av.t_prev = t

  return μ
end
