@doc """
Modal Decay Averaged Artificial Viscosity model, cf. Yu, Hesthaven, 2019.
"""
@with_kw mutable struct MDA <: Abstract_AV
  cmax::Real
  μ::AbstractArray
end


@doc """
Construct `MDA` object based on mesh.
"""
function MDA(prms::Parameters, mesh::Mesh)
  @unpack x = mesh
  μ0 = zeros(Float64, size(x))
  return MDA(prms.cmax, μ0)
end


function compute(u, v, t, mesh::Mesh, eq::Abstract_Eq, av::MDA)
  @unpack cmax = av
  @unpack element, K, L, x = mesh
  @unpack invV, N = element

  # h = L / (K * N)
  h = L / K

  uhat = invV * u  # modal coefficients

  # perfect decay modal coefficients, added to enhance sensitivity to 
  # floating point noise
  vhat = collect(Float64, 1:N).^(-N)
  vhat ./= norm(vhat)

  cellnorm2_u = dg1d.cellwise_inner_product(u, u, mesh)

  maxabs_v = vec(maximum(abs.(v), dims=1))
  μ = @. cmax * h / N * maxabs_v

  # estimate effective decay rate by least square fit of modal coefficients, (3.9)
  # sum_m ( log(|q_m|) - (log(c) - τ log(m)) )^2 = sum_m (y_m - (a + b x_m))^2 -> min
  # log(m) ... grid points in log scale
  rng_N = 1:N
  data_x = @. - log(rng_N)

  for j in 1:K
    # (3.11)
    # only consider coefficients 2:Np (which corresponds to 1:N), because
    # the 1st (0th) one corresponds to a constant mode that is irrelevant
    # for the smoothness esimation.
    mod_uh = @. sqrt( uhat[2:end,j]^2 + cellnorm2_u[j] * vhat^2 )
    # (3.12)
    skylined_mod_uh = [ maximum(mod_uh[min(n,N-1):end]) for n in 1:N ]
    # skylined_mod_uh = zeros(size(mod_uh))
    # for m in 1:N
    #   skylined_mod_uh[m] = maximum(mod_uh[min(m,N-1):end])
    # end

    # data points in log scale
    data_y = @. log(skylined_mod_uh)
    _, _, τ = dg1d.least_sqr_linear(data_x, data_y)

    # if isnan(τ)
    #   μ[j] = 0.0
    # if τ < 1.0
    #   μ[j] *= 1.0
    # elseif τ <= 3.0
    if 1.0 <= τ <= 3.0
      μ[j] *= 1.0 - 0.5 * (τ - 1.0)
    elseif τ > 3.0
      μ[j] *= 0.0
    end
  end

  return μ
end
