export EoS
module EoS


import Parameters: @with_kw, @unpack


@doc """
An EquationOfState (eos) must implement the following functions:

pressure(eos, u..., x) - compute pressure based on state variables (u...)
and grid point x.

TODO Maybe do not export Abstract_EoS, since we encourage to use explicit
types?
"""
abstract type Abstract_EoS end
export Abstract_EoS

abstract type Abstract_Variable end

# TODO Think about dispatching the call operator () with Value types
# and symbols so that one could conviniently run
#   - eos(:p, ρ) to compute p(ρ)
#   - eos(:dpdρ, ρ) to compute dp/dρ(ρ)
# etc.
# Q Why should this be better than defining methods like
#   - p(eos, ρ),
#   - dpdρ(eos, ρ)?
# One advantage would be that no namespace is needed to encapsulate
# p, dpdρ for different EoSs.
# Drawback: Actual syntax looks something like
#   - eos(Val(:p), ρ)
#   - eos(Val(Val(:dpdρ), ρ)


@doc """
Compute pressure from state variables according to equation of state.

Expected to be called with
  pressure(eos, u, x)
where
- eos ... equation of state <: Abstract_EoS
- u   ... (collection) of state variables
- x   ... grid points

# TODO Figure out if x dependency makes even sense.
"""
function pressure end
export pressure

(::Abstract_EoS)(value::Val, vars...) = error("Computation of requested value '$value' is not implemented!")

@doc """
Ideal gas law equation of state

  p(ρ,q,E) = (γ - 1) * (E - q^2 / (2*ρ))

γ ... polytropic index
p ... pressure
ρ ... mass density
q ... momentum density
E ... energy
"""
@with_kw struct IdealGasEoS <: Abstract_EoS
  γ::Float64 = 1.4
end
export IdealGasEoS

pressure(eos::IdealGasEoS, ρ, q, E, x) = (eos.γ - 1.0) * (E - 0.5 * q^2 / ρ)
# (eos::IdealGas)(ρ, q, E, x) = @. (eos.γ - 1.0) * (E - 0.5 * q^2 / ρ)


@doc """
Ideal gas law equation of state (version used for SR Euler Eq)

  p(ρ, ϵ) = (γ - 1) * ρ * ϵ

γ ... polytropic index
ρ ... rest mass density
ϵ ... internal energy

TODO Fiugre out if there is a simple connection between this and the above
version.
"""
@with_kw struct SR_IdealGasEoS <: Abstract_EoS
  γ::Float64 = 1.4
end
export SR_IdealGasEoS

pressure(eos::SR_IdealGasEoS, ρ, ϵ, x) = @. (eos.γ - 1.0) * ρ * ϵ

@doc """
χ = ∂p / ∂ρ
"""
chi(eos::SR_IdealGasEoS, ρ, ϵ, x) = (eos.γ - 1.0) * ϵ

@doc """
κ = ∂p / ∂ϵ
"""
kappa(eos::SR_IdealGasEoS, ρ, ϵ, x) = (eos.γ - 1.0) * ρ


# abstract type Abstract_Variable end
# (var::Abstract_Variable)(arg::Val, ...) =
#   error("Cannot compute ${typeof(var)} for argument $arg, because it is not implemented!")

@doc """
Polytropic equation of state.

  p(ρ) = K ρ^Γ
  ϵ = ρ + p / (Γ - 1)

ρ ... density
K ... polytropic constant
Γ ... polytropic exponent
"""
@with_kw struct PolytropicEoS <: Abstract_EoS
  K::Float64 = 1.0
  Γ::Float64 = 5.0/3.0
end
export PolytropicEoS


# function (eos::PolytropicEoS)(::Val{:p}, ::Val{:ρ}, ρ)
function pressure_from_density(eos::PolytropicEoS, ρ)
# function (eos::PolytropicEoS)(::Val{:p}, ρ)
  @unpack K, Γ = eos
  return K * ρ^Γ
end


function density_from_pressure(eos::PolytropicEoS, p)
# function (eos::PolytropicEoS)(::Val{:ρ}, p)
  @unpack K, Γ = eos
  return (p / K)^(1.0/Γ)
end


function energy_density_from_density(eos::PolytropicEoS, ρ)
# function (eos::PolytropicEoS)(::Val{:ϵ}, ρ)
  @unpack K, Γ = eos
  # p = eos(Val(:p), ρ)
  return ρ + K * ρ^Γ / (Γ - 1.0)
end


function energy_density_from_pressure(eos::PolytropicEoS, p)
  @unpack K, Γ = eos
  return (p/K)^(1.0/Γ) + p / (Γ - 1.0)
end


struct Density <: Abstract_Variable
  var
end

struct EnergyDensity <: Abstract_Variable
  var
end

struct Pressure <: Abstract_Variable
  var
end

function pressure(eos::PolytropicEoS, ρ::Density)
  @unpack K, Γ = eos
  return K * ρ.var^Γ
end

function density(eos::PolytropicEoS, p::Pressure)
  @unpack K, Γ = eos
  return (p.var / K)^(1.0/Γ)
end

function energydensity(eos::PolytropicEoS, ρ::Density)
  @unpack K, Γ = eos
  return ρ.var + K * ρ.var^Γ / (Γ - 1.0)
end

function energydensity(eos::PolytropicEoS, p::Pressure)
  @unpack K, Γ = eos
  return (p.var/K)^(1.0/Γ) + p.var / (Γ - 1.0)
end

end # module


export NewEoS
module NewEoS

using dg1d
using Parameters: @with_kw, @unpack
abstract type Abstract_EoS end

# Check whether this suggestion https://stackoverflow.com/a/26594749
# allows to reuse standard arithmetics
abstract type Abstract_Variable end

# Better syntax
# eos(Pressure, Density(vals...))
# eos(Derivative(Pressure, Density), Density(vals...))

function chain(vars, sep=", ")
  N = length(vars)
  return string([ string(v) * ( (i < N) ? sep : "" ) for (i, v) in enumerate(vars) ]...)
end

# interface methods
function compute(f::Type{<:Abstract_Variable}, eos::Abstract_EoS,
        vars::Abstract_Variable...)
  vars_str = chain([typeof.(vars)...])
  error("Computation of '$(f)' for equation of state " *
        "'$(typeof(eos))' with arguments '$(vars_str)' not implemented!")
end

function compute_derivative(
    f::Type{<:Abstract_Variable}, x::Type{<:Abstract_Variable}, eos::Abstract_EoS,
        vars...)
  vars_str = chain([typeof.(vars)...])
  types_vars = [ typeof(var) for var in vars  ]
  error("Computation of derivative 'd $(f)/d $(x)' for equation of state " *
        "'$(typeof(eos))' with arguments '$(vars_str)' not implemented!")
end


macro Variable(varname, docstr="Documentation missing")
  esc(
    quote
      @doc $docstr
      struct $varname{T} <: dg1d.NewEoS.Abstract_Variable x::T end
      export $varname
    end
  )
end

@Variable Density """
Denoted by ρ.
"""

@Variable Pressure """
Denoted by p.
"""

@Variable InternalEnergy """
Denoted by ϵ.
"""

@Variable TotalEnergy """
Denoted by e and defined as

  e = \rho (1 + \epsilon).

"""

@Variable Enthalpy """
Denoted by h and defined as

  h = 1 + ϵ + p / ρ

"""


# @doc """
# Rest mass density, denoted by \rho.
# """
# struct Density{T} <: Abstract_Variable x::T end
# export Density
#
# @doc """
# Pressure, denoted by p.
# """
# struct Pressure{T} <: Abstract_Variable x::T end
# export Pressure
#
# @doc """
# Internal energy density, denoted by \epsilon.
# """
# struct Energy{T} <: Abstract_Variable x::T end
# export Energy
#
# @doc """
# Total energy density, denoted by e.
# Related to \rho, p, \epsilon by
#
#   e = \rho (1 + \epsilon).
#
# """
# struct TotalEnergy{T} <: Abstract_Variable x::T end
# export TotalEnergy
#
# @doc """
# Enthalpy density, denoted by h.
# Related to \rho, p, \epsilon by
#
#   h = 1 + \epsilon + \frac{p}{\rho}.
#
# """
# struct Enthalpy{T} <: Abstract_Variable x::T end
# export Enthalpy

include("polytropic_eos.jl")
# include("ideal_gas_eos.jl")


end # module
