function evolve(;tend=0.25, save::Bool=false)

  Plots.gr()
  theme(:dark, linewidth=0.5, markerstrokewidth=0, markersize=1.5)

  N = 7
  K = 100
  quadr = :LGL_lumped
  # quadr = :GLGL
  CFL = 0.25
  tstart = 0.0
  # tend = 0.25
  av_method = :MDA
  av_cmax = 1.5
  av_c1 = 0.0
  av_c2 = 0.0
  scheme = :LSERK4
  Rmax = 12.0
  atm_factor = 1e-8
  atm_threshold = 100.0
  tov_id_filename = "TOV.jld2"


  element = SpectralElement(N, quadr)
  mesh = dg1d.Mesh(element, K, -Rmax, Rmax, CFL)
  av_prms = AV.Parameters(av_method, av_cmax, tstart, av_c1, av_c2)
  av = AV.make_AV(av_prms, mesh)

  rhs = GRHD.Rhs(mesh, av)
  ρatm = read_tov_initialdata!(rhs, tov_id_filename,
                                atm_factor=atm_factor, atm_threshold=atm_threshold)
  # ρatm = 1e-11
  # read_atmosphere_initialdata!(rhs, tov_id_filename, ρatm=ρatm)

  av.μ .= 0.0

  components = [ ScaledConservatives ]
  @unpack K = mesh
  @unpack Npts = element
  sv_u = StateVector(Float64, K*Npts, components)
  sv_du = StateVector(Float64, K*Npts, components)

  sv_u.data .= vec(rhs.scons)

  outputdir = dg1d.get_output_dir()
  # prms_weno = dg1d.Parameters_WENO()
  # weno_rec = dg1d.WENOReconstruction(prms_weno, mesh)
  # f_rhs(du, u, t) = rhs_weno!(rhs, du, u, t, components, ρatm, atm_threshold,
  #                                  weno_rec)
  bernstein_rec = dg1d.BernsteinReconstruction(element.z)
  f_rhs(du, u, t) = rhs_bernstein!(rhs, du, u, t, components, ρatm, atm_threshold,
                                   bernstein_rec)
  # f_rhs(du, u, t) = rhs_ldg!(rhs, du, u, t, components, ρatm, atm_threshold)

  function fn_dt(u, t)
    sv_u = StateVector(u, components)
    rhs.scons = get_component(1, sv_u)
    compute_conservatives!(rhs.cons, rhs.scons, rhs.mfs, mesh)
    compute_primitives!(rhs, ρatm=ρatm, atm_threshold=atm_threshold)
    vmax = ones(Npts * K)
    return AV.timestep(u, vmax, mesh, av)
  end

  hist =  [ (tstart, deepcopy(rhs)) ]
  # save_steps = 40
  # save_dt = (tend - tstart) / save_steps
  save_dt = 0.5
  save_counter = 0

  intrp_order = 1
  intrp_element = dg1d.SpectralElement(intrp_order, quadr)
  intrp_from  = mesh.element.z
  intrp_to    = intrp_element.z
  # intrp_mat   = dg1d.barycentric_interp_matrix(intrp_from, intrp_to, intrp_element.bw)
  intrp_mat   = dg1d.lagrange_interp_matrix(intrp_from, intrp_to)
  # post hook run after each full RK step
  function hook(u, t)
    @unpack x = mesh
    sv_u = StateVector(u, components)
    vr = reshape(rhs.prim.vr, (Npts, K))
    D = reshape(rhs.scons.hatD, (Npts, K))
    τ = reshape(rhs.scons.hatτ, (Npts, K))
    vmax = ones(size(vr))
    # μ_vr = AV.compute(vr, vmax, t, mesh, av)
    μ_D = AV.compute(D, vmax, t, mesh, av)
    # μ_τ = AV.compute(τ, vmax, t, mesh, av)
    # av.μ .= dg1d.smooth_C0(μ_vr, intrp_mat)
    av.μ .= dg1d.smooth_C0(μ_D, intrp_mat)
    # av.μ .= dg1d.smooth_C0(μ_τ, intrp_mat)
    
    # @debug_plot(rhs.av_vars, mesh)
    # @debug_plot(rhs.prim, mesh)

    if (save_counter * save_dt < t)
      push!(hist, (t, deepcopy(rhs)))
      save_counter += 1
    end

    return true
  end

  tspan = (tstart, tend)
  u_0 = vec(sv_u)
  dt = fn_dt(u_0, tstart)

  hook(u_0, tstart)
  u_num = dg1d.evolve_ERK(f_rhs, u_0, dt, tspan, scheme, post_hook=hook)

  # sv_u = StateVector(u_num, components)
  # scons_final = get_component(1, sv_u)
  # compute_conservatives!(rhs.cons, scons_final, rhs.mfs, mesh)
  # compute_primitives!(rhs, ρatm=ρatm, atm_threshold=atm_threshold)
  # compute_primitives!(rhs.prim, rhs.cons, rhs.mfs, rhs.mesh, rhs.eos,
  #                     ρatm=ρatm, atm_threshold=atm_threshold)
  # @debug_plot(rhs.prim, mesh)

  if save
    output_dir = dg1d.get_output_dir()
    filename = joinpath(output_dir, "bernstein_n4/rhs.jld")
    jldsave(filename; hist)
  end

  return
end


function analyze()

  output_dir = dg1d.get_output_dir()
  filename = "bernstein_n4/rhs.jld"
  filename = joinpath(output_dir, filename)
  plotname(basename, nr) = joinpath(output_dir, "$(basename).$(lpad(nr,3,'0')).png")

  hist = nothing
  jldopen(filename, "r") do file
    hist = read(file, "hist")
  end

  prog = Progress(length(hist))

  for (i, h) in enumerate(hist)
    t, rhs = h
    plt = dg1d.plot(rhs.cons, rhs.mesh)
    Plots.plot!(plt, title="T = $t")
    pltname = plotname("conservatives", i)
    png(pltname)
    plt = dg1d.plot(rhs.scons, rhs.mesh)
    Plots.plot!(plt, title="T = $t")
    pltname = plotname("scaledconservatives", i)
    png(pltname)
    savefig(plt, pltname)
    plt = dg1d.plot(rhs.prim, rhs.mesh)
    Plots.plot!(plt, title="T = $t")
    pltname = plotname("primitives", i)
    png(pltname)
    next!(prog)
  end

end
