function schwarzschild_solution_isotropic_coordinates(R, M)
  Rmin = minimum(R)
  @assert 2 * Rmin >= M

  A = @. (1 - 0.5 * M / R) / (1 + 0.5 * M / R)
  B = @. (1 + 0.5 * M / R)^2

  return A, B
end


function interpolate(x, x_ref, y_ref)
  x_intrp = LinearInterpolation(x_ref, y_ref, extrapolation_bc=Line())
  y = [ x_intrp(xi) for xi in x ] 
  return y
end


function mirror(A, exclude_bdry::Bool)
  if exclude_bdry
    return cat(A[end:-1:2], A, dims=1)
  else
    return cat(A[end:-1:1], A, dims=1)
  end
end


function reflect(A, exclude_bdry::Bool)
  if exclude_bdry
    return cat(-A[end:-1:2], A, dims=1)
  else
    return cat(-A[end:-1:1], A, dims=1)
  end
end


function read_tov_initialdata!(rhs::Rhs, filename;
    atm_factor=1e-8, atm_threshold=100.0)

  @unpack mesh = rhs
  @unpack Npts = mesh.element
  # DG coordinates
  intrp_R = vec(mesh.x)
  NN = length(intrp_R)
  
  # determine if grid contains r = 0 point
  R_ismirrored = intrp_R[1] ≈ -intrp_R[end]
  R_includes_origin = (
    !R_ismirrored # LGL grids always contain bdry points
    || ( 
      iseven(mesh.K) ||  # origin falls at interface between two cells
      (isodd(mesh.K) && isodd(mesh.element.Npts)) # origin sits in the center 
      # of a cell with a point in the center
    )
  )
  R_origin_isdoubled = (
    R_ismirrored && # only happens for mirrored domains
    iseven(mesh.K)  # and origin falls at interface between two cells
  )
  
  # on a mirrored grid we only do interpolation on one half and then
  # mirror/reflect onto the other
  if R_ismirrored
    intrp_R = intrp_R[Int(floor(NN/2))+1:end]
  end
  # if r = 0 only covered by one point we have to exclude the boundary
  # point when mirroring/refelcting back
  exclude_bdry = !R_origin_isdoubled && R_includes_origin
  Rmax = maximum(intrp_R)

  # read TOV result from file
  # only contains solution in star interior
  output_dir = dg1d.get_output_dir()
  sol, prms = JLD2.load(joinpath(output_dir, filename), "sol", "prms")
  ϕint = sol[TOV.idx_ϕ,:]
  Rint = sol[TOV.idx_R,:]
  rint = sol[TOV.idx_r,:]
  @unpack K, Γ, ρc, M = prms
  Rs = Rint[end]

  mat_intrp_R = reshape(intrp_R, (Npts, Int64(mesh.K/2)))
  _, Rs_idx = findmin(abs.(mat_intrp_R.-Rs))
  n, k = Tuple(Rs_idx)
  @info "TOV star surface located at R = $Rs"
  @info "(n,k) = ($n, $k)"

  #### setup metric functions A, B and derivatives in isotropic coordinates

  # star interior solution
  Aint = exp.(ϕint)
  # B := r / R, needs special treatment of coordinate singularity.
  Bint = [ (ri == 0) ? 1.0 : ri / Ri for (ri, Ri) in zip(rint, Rint) ]
  # use linear interpolation to obtain value at r = R = 0.
  Bint[1] = (Bint[3] - Bint[2]) / (Rint[3] - Rint[2]) * (0.0 - Rint[2]) + Bint[2]

  # derivatives
  dAint = dg1d.FD.central_dx(Aint, Rint)
  dBint = dg1d.FD.central_dx(Bint, Rint)

  # star exterior solution =^= Schwarzschild solution
  Npts_ext = 1000 # number of points used for interpolation of exterior solution
  Rext = collect(LinRange(Rs, Rmax, Npts_ext))
  Aext, Bext = schwarzschild_solution_isotropic_coordinates(Rext, M)

  # merge interior and exterior solutions
  # note: exclude surface point in one of the arrays when merging
  R = cat(Rint, Rext[2:end], dims=1)
  A = cat(Aint, Aext[2:end], dims=1)
  B = cat(Bint, Bext[2:end], dims=1)

  # derivatives
  dA = dg1d.FD.central_dx(A, R)
  dB = dg1d.FD.central_dx(B, R)

  intrp_A   = interpolate(intrp_R[:], R, A)
  intrp_B   = interpolate(intrp_R[:], R, B)
  intrp_A_r  = interpolate(intrp_R[:], R, dA)
  intrp_B_r  = interpolate(intrp_R[:], R, dB)

  if R_ismirrored
    intrp_A    = mirror(intrp_A, exclude_bdry)
    intrp_B    = mirror(intrp_B, exclude_bdry)
    intrp_A_r  = reflect(intrp_A_r, exclude_bdry)
    intrp_B_r  = reflect(intrp_B_r, exclude_bdry)
  end
  
  # populate metric fns in rhs by interpolating above data linearly
  @unpack mfs = rhs
  mfs.A   .= intrp_A
  mfs.B   .= intrp_B
  mfs.A_r .= intrp_A_r
  mfs.B_r .= intrp_B_r
  @assert isnan(mfs) == false
  @assert isinf(mfs) == false

  # @debug_plot(mfs, mesh)

  ### setup primitive variables
  
  rhs.eos = PolytropicEoS(K = prms.K, Γ = prms.Γ)
  @unpack eos = rhs

  # star interior solution
  pint = sol[TOV.idx_p,:]
  pint[@.(pint < 0)] .= 0.0
  # apply atmosphere treatment
  pmax = maximum(pint)
  ρmax = NewEoS.compute(Density, eos, Pressure(pmax))
  ρatm = atm_factor * ρmax
  patm = NewEoS.compute(Pressure, eos, Density(ρatm))
  ρint = NewEoS.compute.(Density, eos, Pressure.(pint))
  idxatm_treatment = @. ρint < ρatm * atm_threshold
  pint[idxatm_treatment] .= patm
  pint[idxatm_treatment] .= ρatm
  vint = zeros(size(Rint))
  ϵint = NewEoS.compute.(InternalEnergy, eos, Pressure.(pint), Density.(ρint))

  # star exterior data
  ρext = ρatm * ones(size(Rext))
  pext = NewEoS.compute.(Pressure, eos, Density.(ρext))
  vext = zeros(size(Rext))
  ϵext = NewEoS.compute.(InternalEnergy, eos, Pressure.(pext), Density.(ρext))

  # merge interior and exterior solutions
  p = cat(pint, pext[2:end], dims=1)
  ρ = cat(ρint, ρext[2:end], dims=1)
  vr = cat(vint, vext[2:end], dims=1)
  ϵ = cat(ϵint, ϵext[2:end], dims=1)

  intrp_p = interpolate(intrp_R[:], R, p)
  intrp_ρ = interpolate(intrp_R[:], R, ρ)
  intrp_vr = interpolate(intrp_R[:], R, vr)
  intrp_ϵ = interpolate(intrp_R[:], R, ϵ)

  if R_ismirrored
    intrp_p = mirror(intrp_p, exclude_bdry)
    intrp_ρ = mirror(intrp_ρ, exclude_bdry)
    intrp_vr = reflect(intrp_vr, exclude_bdry)
    intrp_ϵ = mirror(intrp_ϵ, exclude_bdry)
  end

  # populate primitive variables with extracted data
  @unpack prim = rhs
  prim.p .= intrp_p
  prim.ρ .= intrp_ρ
  prim.vr .= intrp_vr
  prim.ϵ .= intrp_ϵ
  @assert isnan(prim) == false
  @assert isinf(prim) == false

  # @debug_plot(prim, mesh)

  @unpack cons = rhs
  compute_conservatives!(cons, prim, mfs)
  @assert isnan(cons) == false
  @assert isinf(cons) == false

  # @debug_plot(cons, mesh)

  @unpack scons = rhs
  compute_scaledconservatives!(scons, cons, mfs, mesh)
  # @debug_plot(scons, mesh)
  @assert isnan(cons) == false
  @assert isinf(cons) == false

  # compute_primitives!(rhs.prim, rhs.cons, rhs.mfs, rhs.mesh, rhs.eos,
  #                     ρatm=ρatm, atm_threshold=atm_threshold)

  return ρatm
end



function read_atmosphere_initialdata!(rhs::Rhs, filename;
    ρatm=1e-11)

  @unpack mesh = rhs
  @unpack Npts = mesh.element
  # DG coordinates
  R = vec(mesh.x)
  
  #### setup metric functions A, B and derivatives in isotropic coordinates

  # star interior solution
  A = ones(size(R))
  B = ones(size(R))

  # derivatives
  A_r = zeros(size(R))
  B_r = zeros(size(R))

  # populate metric fns in rhs by interpolating above data linearly
  @unpack mfs = rhs
  mfs.A   .= A
  mfs.B   .= B
  mfs.A_r .= A_r
  mfs.B_r .= B_r
  @assert isnan(mfs) == false
  @assert isinf(mfs) == false

  ### setup primitive variables
  
  # equation of state for atmosphere
  rhs.eos = PolytropicEoS(K = 100.0, Γ = 2.0)
  @unpack eos = rhs

  ρ = ρatm * ones(size(R))
  p = NewEoS.compute.(Pressure, eos, Density.(ρ))
  vr = zeros(size(R))
  ϵ = NewEoS.compute.(InternalEnergy, eos, Pressure.(p), Density.(ρ))

  # populate primitive variables with extracted data
  @unpack prim = rhs
  prim.p .= p
  prim.ρ .= ρ
  prim.vr .= vr
  prim.ϵ .= ϵ
  @assert isnan(prim) == false
  @assert isinf(prim) == false

  @unpack cons = rhs
  compute_conservatives!(cons, prim, mfs)
  @assert isnan(cons) == false
  @assert isinf(cons) == false

  @unpack scons = rhs
  compute_scaledconservatives!(scons, cons, mfs, mesh)
  @assert isnan(cons) == false
  @assert isinf(cons) == false

  return ρatm
end
