function find_troubled_cells(fn, var, mesh)
  idxs = findall(fn, var)
  cells = [ Tuple(idx)[2] for idx in idxs ]
  unique!(cells)
  return cells
end


function reconstruct_scaledconservatives!(rhs::GRHD.Rhs, 
    bernstein_rec::dg1d.BernsteinReconstruction, ρatm)

  @unpack mesh = rhs
  @unpack Npts = mesh.element
  @unpack K = mesh
  shape = (Npts, K)

  @unpack μ = rhs.av
  # troubled cells based on where visocisty is non-vanishing
  # trbld_cells = find_troubled_cells(!iszero, μ, mesh)
  # manually add surface cells and bdrys for now
  trbld_cells = []
  push!(trbld_cells, 1)
  push!(trbld_cells, K)
  push!(trbld_cells, 86:87)
  push!(trbld_cells, 15:16)
  # append!(trbld_cells, collect(83:87))
  # append!(trbld_cells, collect(15:19))
  unique!(trbld_cells)

  # use only Bernstein reconstruction for now
  w = 0.2 # convex weighting

  # atmosphere values
  patm = NewEoS.compute(Pressure, rhs.eos, Density(ρatm))
  hatm = NewEoS.compute(Enthalpy, rhs.eos, Pressure(patm), Density(ρatm))
  Watm = 1.0 # stationary atmosphere

  mat_B = reshape(rhs.mfs.B, shape)
  mat_r = reshape(rhs.mesh.x, shape)

  # apply reconstruction to scaled conservatives
  var = rhs.scons
  for (i,(var,varname)) in enumerate(zip(fields(var), fieldnames(typeof(var))))
    mat_var = reshape(var, shape)
    for cell in trbld_cells
      trbld_var = view(mat_var, 1:Npts, cell)
      trbld_B = view(mat_B, 1:Npts, cell)
      trbld_r = view(mat_r, 1:Npts, cell)
      # enforce lower bounds on hatD and hatτ through a stationary atmosphere
      if varname === :hatD
        trbld_min_hatD = @. ρatm * Watm * trbld_B^3 * trbld_r
        if minimum(trbld_r) > 0
          dg1d.reconstruct!(trbld_var, w, bernstein_rec, m=trbld_min_hatD)
        else
          dg1d.reconstruct!(trbld_var, w, bernstein_rec, M=trbld_min_hatD)
        end
      elseif varname === :hatτ
        trbld_min_hatτ = @. (ρatm * hatm * Watm^2 - patm - ρatm * Watm) * trbld_B^3 * trbld_r
        if minimum(trbld_r) > 0
          dg1d.reconstruct!(trbld_var, w, bernstein_rec, m=trbld_min_hatτ)
        else
          dg1d.reconstruct!(trbld_var, w, bernstein_rec, M=trbld_min_hatτ)
        end
      else
        # remaining variables don't need bounds
        dg1d.reconstruct!(trbld_var, w, bernstein_rec)
      end
    end
  end

end


function reconstruct_conservatives!(rhs::GRHD.Rhs, bernstein_rec::dg1d.BernsteinReconstruction,
    ρatm; reconstruction_variables::Vector{Symbol}=Symbol[], impose_bounds::Bool=true)

  variables = rhs.cons
  vars = fieldnames(typeof(variables))
  @assert all(in.(reconstruction_variables, Ref(vars))) "Provided variables '$vars' not available in `$vars`"

  @unpack mesh = rhs
  @unpack Npts = mesh.element
  @unpack K = mesh
  shape = (Npts, K)

  @unpack μ = rhs.av
  # troubled cells based on where visocisty is non-vanishing
  # trbld_cells = find_troubled_cells(!iszero, μ, mesh)
  # manually add surface cells and bdrys for now
  trbld_cells = []
  push!(trbld_cells, 1)
  push!(trbld_cells, K)
  push!(trbld_cells, 83)
  push!(trbld_cells, 18)
  # append!(trbld_cells, collect(84:87))
  # append!(trbld_cells, collect(15:18))
  unique!(trbld_cells)

  # use only Bernstein reconstruction for now
  w = 0.0 # convex weighting

  # atmosphere values
  patm = NewEoS.compute(Pressure, rhs.eos, Density(ρatm))
  hatm = NewEoS.compute(Enthalpy, rhs.eos, Pressure(patm), Density(ρatm))
  Watm = 1.0 # stationary atmosphere

  # apply reconstruction to conservatives
  for rv in reconstruction_variables
    var = getfield(variables, rv)
    mat_var = reshape(var, shape)
    for cell in trbld_cells
      trbld_var = view(mat_var, 1:Npts, cell)
      if impose_bounds
        if :D in vars
          min_D = ρatm * Watm
          dg1d.reconstruct!(trbld_var, w, bernstein_rec, m=min_D)
        elseif :τ in vars
          min_τ = (ρatm * hatm * Watm^2 - patm - ρatm * Watm)
          dg1d.reconstruct!(trbld_var, w, bernstein_rec, m=min_τ)
        elseif :Sr in vars
          dg1d.reconstruct!(trbld_var, w, bernstein_rec)
        end
      else
        dg1d.reconstruct!(trbld_var, w, bernstein_rec)
      end
    end
  end

end



function reconstruct_primitives!(rhs::GRHD.Rhs, bernstein_rec::dg1d.BernsteinReconstruction,
    ρatm; reconstruction_variables::Vector{Symbol}=[:vr], impose_bounds::Bool=true)

  variables = rhs.prim
  vars = fieldnames(typeof(variables))
  @assert all(in.(reconstruction_variables, Ref(vars))) "Provided variables '$vars' not available in `$vars`"

  @unpack mesh = rhs
  @unpack Npts = mesh.element
  @unpack K = mesh
  shape = (Npts, K)

  # troubled cells based on where visocisty is non-vanishing
  @unpack μ = rhs.av
  indicator = dg1d.TroubledCells.ArtificialViscosityIndicator(0.8)
  μnorm = sum(μ, dims=1) / Npts
  μnorm_max = maximum(μnorm)
  if μnorm_max > 0.0 
    μnorm ./= μnorm_max
  end

  trbld_cells = dg1d.TroubledCells.find_troubled_cells(indicator, μnorm, mesh)
  push!(trbld_cells, 1)
  push!(trbld_cells, K)
  unique!(trbld_cells)

  # use only Bernstein reconstruction for now
  w = 0.0 # convex weighting

  # atmosphere values
  patm = NewEoS.compute(Pressure, rhs.eos, Density(ρatm))
  ϵatm = NewEoS.compute(InternalEnergy, rhs.eos, Pressure(patm), Density(ρatm))
  vmax = 1.0 - 1e-3

  # apply reconstruction to selected variables `vars`
  for rv in reconstruction_variables
    var = getfield(variables, rv)
    mat_var = reshape(var, shape)
    for cell in trbld_cells
      trbld_var = view(mat_var, 1:Npts, cell)
      if impose_bounds
        if :ρ in vars
          dg1d.reconstruct!(trbld_var, w, bernstein_rec, m=ρatm)
        elseif :p in vars
          dg1d.reconstruct!(trbld_var, w, bernstein_rec, m=patm)
        elseif :ϵ in vars
          dg1d.reconstruct!(trbld_var, w, bernstein_rec, m=ϵatm)
        elseif :vr in vars
          dg1d.reconstruct!(trbld_var, w, bernstein_rec, m=-vmax, M=vmax)
        end
      else
        dg1d.reconstruct!(trbld_var, w, bernstein_rec)
      end
    end
  end

end


function rhs_bernstein!(rhs::Rhs, dU, U, t, 
    components, ρatm, atm_threshold, 
    bernstein_rec::dg1d.BernsteinReconstruction)

  @unpack mesh = rhs
  @unpack K = mesh
  @unpack Npts = mesh.element
  shape = (Npts, K)

  Rs = 8.125144892166112
  idx_lhs_surface = dg1d.locate_point(-Rs, mesh)
  idx_rhs_surface = dg1d.locate_point(Rs, mesh)

  stencile_width = 4
  stencile(i) = (i-stencile_width):(i+stencile_width)

  rng_lhs_surface = (:, stencile(idx_lhs_surface[2]))
  rng_rhs_surface = (:, stencile(idx_lhs_surface[2]))

  ### interpret inputs
  sv_u = StateVector(U, components)
  sv_du = StateVector(dU, components)
  rhs.scons = get_component(1, sv_u)
  d_scons = get_component(1, sv_du)

  display(dg1d.Segment(rhs.prim, mesh, rng_lhs_surface...))
  @debug_plot(dg1d.Segment(rhs.prim, mesh, rng_lhs_surface...), mesh)
  compute_conservatives!(rhs.cons, rhs.scons, rhs.mfs, mesh)
  compute_primitives!(rhs, ρatm=ρatm, atm_threshold=atm_threshold)
  reconstruct_primitives!(rhs, bernstein_rec, ρatm, reconstruction_variables=[:vr],
    impose_bounds=false)
  compute_conservatives!(rhs.cons, rhs.prim, rhs.mfs)
  compute_scaledconservatives!(rhs.scons, rhs.cons, rhs.mfs, mesh)
  display(dg1d.Segment(rhs.prim, mesh, rng_lhs_surface...))
  @debug_plot(dg1d.Segment(rhs.prim, mesh, rng_lhs_surface...), mesh)
  # display(dg1d.Segment(rhs.prim, mesh, rng_lhs_surface...))
  # @debug_plot(dg1d.Segment(rhs.prim, mesh, rng_lhs_surface...), mesh)

  compute_fluxes!(rhs.fluxes, rhs.prim, rhs.mfs, mesh.x[:])
  compute_sources!(rhs.sources, rhs.prim, rhs.mfs, mesh.x[:];
    ρatm=ρatm, atm_threshold=atm_threshold)

  ### actual RHS computation
  # solve conservation law: ∂t u + ∂x f(u) + ∂x μ q = s

  @unpack fD, fSr, fτ = rhs.fluxes
  @unpack hatD, hatSr, hatτ = rhs.scons

  # numerical fluxes
  # For now we use speed of light as upper bound for lax friedrich flux
  max_v = 1.0
  # TODO implement estimation of max wave speed, requires characteristics
  # max_v = maxabs_v(eq, ...)

  intf_hatD = dg1d.to_interface(reshape(hatD, shape), mesh)
  intf_hatSr = dg1d.to_interface(reshape(hatSr, shape), mesh)
  intf_hatτ = dg1d.to_interface(reshape(hatτ, shape), mesh)
  intf_fD = dg1d.to_interface(reshape(fD, shape), mesh)
  intf_fSr = dg1d.to_interface(reshape(fSr, shape), mesh)
  intf_fτ = dg1d.to_interface(reshape(fτ, shape), mesh)
  nfD_lhs, nfD_rhs = dg1d.lax_friedrich_flux(intf_fD, intf_hatD, max_v)
  nfSr_lhs, nfSr_rhs = dg1d.lax_friedrich_flux(intf_fSr, intf_hatSr, max_v)
  nfτ_lhs, nfτ_rhs = dg1d.lax_friedrich_flux(intf_fτ, intf_hatτ, max_v)
  # nfD_lhs, nfD_rhs = dg1d.central_flux(intf_fD)
  # nfSr_lhs, nfSr_rhs = dg1d.central_flux(intf_fSr)
  # nfτ_lhs, nfτ_rhs = dg1d.central_flux(intf_fτ)

  # # Set boundary conditions
  # Not needed for now, because we try periodic BCs first.
  # nfD_lhs, nfD_rhs
  # nfSr_lhs, nfSr_rhs
  # nfτ_lhs, nfτ_rhs

  dg1d.new_rhs_weak!(d_scons.hatD, fD, rhs.sources.sD, nfD_lhs, nfD_rhs, mesh)
  dg1d.new_rhs_weak!(d_scons.hatSr, fSr, rhs.sources.sSr, nfSr_lhs, nfSr_rhs, mesh)
  dg1d.new_rhs_weak!(d_scons.hatτ, fτ, rhs.sources.sτ, nfτ_lhs, nfτ_rhs, mesh)
  # display(dg1d.Segment(rhs.prim, mesh, rng...))
  # readline()
  # @debug_plot(dg1d.Segment(rhs.prim, mesh, rng...), mesh, (fn=(u)->abs.(u)+1e-20, yscale=:log10))
  # @debug_plot(dg1d.Segment(rhs.prim, mesh, rng...), mesh)
  # @debug_plot(dg1d.Segment(rhs.scons, mesh, rng...), mesh)
  # @debug_plot(dg1d.Segment(rhs.sources, mesh, rng...), mesh)
  # @debug_plot(dg1d.Segment(rhs.fluxes, mesh, rng...), mesh)

  # dg1d.plot_interfaces([nfD_lhs, nfSr_lhs, nfτ_lhs], [nfD_rhs, nfSr_rhs, nfτ_rhs], mesh;
  #     xlims=(mesh.x[1,rng[2].start], mesh.x[end,rng[2].stop]))
  # readline()

  # display(dg1d.Segment(d_scons, mesh, rng...))
  # mat_hatSr = reshape(d_scons.hatSr, shape)
  # mat_hatSr[:,83:84] .= 0
  # mat_hatSr[:,17:18] .= 0
  # @debug_plot(dg1d.Segment(d_scons, mesh, rng...), mesh)
  # @debug_plot(d_scons, mesh)

  # needed to mitigate effects caused by imposing periodic boundary conditions wrongly
  d_scons.hatD[1] = 0.0; d_scons.hatD[end] = 0.0
  d_scons.hatSr[1] = 0.0; d_scons.hatSr[end] = 0.0
  d_scons.hatτ[1] = 0.0; d_scons.hatτ[end] = 0.0

  return
end
