function reconstruct_scaledconservatives!(rhs::GRHD.Rhs, 
    weno::dg1d.WENOReconstruction, ρatm)

  @unpack mesh = rhs
  @unpack Npts = mesh.element
  @unpack K = mesh
  shape = (Npts, K)

  indicator = dg1d.TroubledCells.MinmodIndicator()
  
  trbld_cells = Int64[]
  for var in fields(rhs.scons)
    mat_var = reshape(var, shape)
    avgs = dg1d.cellwise_average(mat_var, mesh)
    append!(trbld_cells, dg1d.TroubledCells.find_troubled_cells(indicator, var, avgs, mesh))
  end

  append!(trbld_cells, collect(83:87))
  append!(trbld_cells, collect(14:18))
  unique!(trbld_cells)
  sort!(trbld_cells)

  for var in fields(rhs.scons)
    dg1d.reconstruct!(var, trbld_cells, weno)
  end

end


# function reconstruct_conservatives!(rhs::GRHD.Rhs, weno::dg1d.WENOReconstruction,
#     ρatm)
#
#   @unpack mesh = rhs
#   @unpack Npts = mesh.element
#   @unpack K = mesh
#   shape = (Npts, K)
#
#   @unpack μ = rhs.av
#   # troubled cells based on where visocisty is non-vanishing
#   # trbld_cells = find_troubled_cells(!iszero, μ, mesh)
#   # manually add surface cells and bdrys for now
#   trbld_cells = []
#   # push!(trbld_cells, 85)
#   # push!(trbld_cells, 16)
#   append!(trbld_cells, collect(83:87))
#   append!(trbld_cells, collect(14:18))
#   unique!(trbld_cells)
#
#   # use only Bernstein reconstruction for now
#   w = 0.5 # convex weighting
#
#   # atmosphere values
#   patm = NewEoS.compute(Pressure, rhs.eos, Density(ρatm))
#   hatm = NewEoS.compute(Enthalpy, rhs.eos, Pressure(patm), Density(ρatm))
#   Watm = 1.0 # stationary atmosphere
#
#   # apply reconstruction to scaled conservatives
#   var = rhs.cons
#   for (i,(var,varname)) in enumerate(zip(fields(var), fieldnames(typeof(var))))
#     mat_var = reshape(var, shape)
#     for cell in trbld_cells
#       trbld_var = view(mat_var, 1:Npts, cell)
#       # enforce lower bounds on hatD and hatτ through a stationary atmosphere
#       if varname === :D
#         trbld_min_D = @. ρatm * Watm
#         dg1d.reconstruct!(trbld_var, w, weno, m=trbld_min_D)
#       elseif varname === :τ
#         # trbld_min_τ = @. (ρatm * hatm * Watm^2 - patm - ρatm * Watm)
#         trbld_min_τ = @. (ρatm * hatm * Watm^2 - patm - ρatm * Watm)
#         dg1d.reconstruct!(trbld_var, w, weno, m=trbld_min_τ)
#       else
#         # remaining variables don't need bounds
#         dg1d.reconstruct!(trbld_var, w, weno)
#       end
#     end
#   end
#
# end


function rhs_weno!(rhs::Rhs, dU, U, t, 
    components, ρatm, atm_threshold, 
    weno::dg1d.WENOReconstruction)

  @unpack mesh = rhs
  @unpack K = mesh
  @unpack Npts = mesh.element
  shape = (Npts, K)

  # rng = (:, K-4:K)
  # rng = (:, 82:86)
  rng = (:, 80:88)
  # drng = (:, 83:85)
  # zoom_rng = (:,84:86)
  zoom_rng = rng

  ### interpret inputs
  sv_u = StateVector(U, components)
  sv_du = StateVector(dU, components)
  rhs.scons = get_component(1, sv_u)
  d_scons = get_component(1, sv_du)

  # @debug_plot(dg1d.Segment(rhs.cons, mesh, rng...), mesh, (fn=(u)->abs.(u), yscale=:log10))
  reconstruct_scaledconservatives!(rhs, weno, ρatm)
  compute_conservatives!(rhs.cons, rhs.scons, rhs.mfs, mesh)
  # @debug_plot(dg1d.Segment(rhs.cons, mesh, rng...), mesh, (fn=(u)->abs.(u), yscale=:log10))
  compute_primitives!(rhs, ρatm=ρatm, atm_threshold=atm_threshold)
  compute_scaledconservatives!(rhs.scons, rhs.cons, rhs.mfs, mesh)

  compute_fluxes!(rhs.fluxes, rhs.prim, rhs.mfs, mesh.x[:])
  compute_sources!(rhs.sources, rhs.prim, rhs.mfs, mesh.x[:];
    ρatm=ρatm, atm_threshold=atm_threshold)

  ### actual RHS computation
  # solve conservation law: ∂t u + ∂x f(u) + ∂x μ q = s

  @unpack fD, fSr, fτ = rhs.fluxes
  @unpack hatD, hatSr, hatτ = rhs.scons

  # numerical fluxes
  # For now we use speed of light as upper bound for lax friedrich flux
  max_v = 1.0
  # TODO implement estimation of max wave speed, requires characteristics
  # max_v = maxabs_v(eq, ...)

  intf_hatD = dg1d.to_interface(reshape(hatD, shape), mesh)
  intf_hatSr = dg1d.to_interface(reshape(hatSr, shape), mesh)
  intf_hatτ = dg1d.to_interface(reshape(hatτ, shape), mesh)
  intf_fD = dg1d.to_interface(reshape(fD, shape), mesh)
  intf_fSr = dg1d.to_interface(reshape(fSr, shape), mesh)
  intf_fτ = dg1d.to_interface(reshape(fτ, shape), mesh)
  nfD_lhs, nfD_rhs = dg1d.lax_friedrich_flux(intf_fD, intf_hatD, max_v)
  nfSr_lhs, nfSr_rhs = dg1d.lax_friedrich_flux(intf_fSr, intf_hatSr, max_v)
  nfτ_lhs, nfτ_rhs = dg1d.lax_friedrich_flux(intf_fτ, intf_hatτ, max_v)
  # nfD_lhs, nfD_rhs = dg1d.central_flux(intf_fD)
  # nfSr_lhs, nfSr_rhs = dg1d.central_flux(intf_fSr)
  # nfτ_lhs, nfτ_rhs = dg1d.central_flux(intf_fτ)

  # # Set boundary conditions
  # Not needed for now, because we try periodic BCs first.
  # nfD_lhs, nfD_rhs
  # nfSr_lhs, nfSr_rhs
  # nfτ_lhs, nfτ_rhs

  dg1d.new_rhs_weak!(d_scons.hatD, fD, rhs.sources.sD, nfD_lhs, nfD_rhs, mesh)
  dg1d.new_rhs_weak!(d_scons.hatSr, fSr, rhs.sources.sSr, nfSr_lhs, nfSr_rhs, mesh)
  dg1d.new_rhs_weak!(d_scons.hatτ, fτ, rhs.sources.sτ, nfτ_lhs, nfτ_rhs, mesh)
  # display(dg1d.Segment(rhs.prim, mesh, rng...))
  # @debug_plot(dg1d.Segment(rhs.prim, mesh, rng...), mesh, (fn=(u)->abs.(u)+1e-20, yscale=:log10))
  # @debug_plot(dg1d.Segment(rhs.prim, mesh, rng...), mesh)
  # @debug_plot(dg1d.Segment(rhs.scons, mesh, rng...), mesh)
  # @debug_plot(dg1d.Segment(rhs.sources, mesh, rng...), mesh)
  # @debug_plot(dg1d.Segment(rhs.fluxes, mesh, rng...), mesh)

  # dg1d.plot_interfaces([nfD_lhs, nfSr_lhs, nfτ_lhs], [nfD_rhs, nfSr_rhs, nfτ_rhs], mesh;
  #     xlims=(mesh.x[1,rng[2].start], mesh.x[end,rng[2].stop]))
  # readline()

  # @debug_plot(dg1d.Segment(d_scons, mesh, rng...), mesh)

  # needed to mitigate effects caused by imposing periodic boundary conditions wrongly
  d_scons.hatD[1] = 0.0; d_scons.hatD[end] = 0.0
  d_scons.hatSr[1] = 0.0; d_scons.hatSr[end] = 0.0
  d_scons.hatτ[1] = 0.0; d_scons.hatτ[end] = 0.0

  return
end
