@with_kw struct Eq <: Abstract_SystemOfEqs
  eos::EoS.SR_IdealGasEoS
  prms::dg1d.NlSolveParameters = dg1d.NlSolveParameters()
end


# TODO Think about moving all of the formulae below into a module to 
# avoid name clashes with ρ, v, ϵ, etc.


module Formulae
  # auxilary variables: W, h, s
  Wprim(ρ, v, ϵ, p) = 1.0 / sqrt(1.0 - v^2) # in terms of primitive variables
  Wcons(D, S, τ, p) = (τ + p + D) / sqrt((τ + p + D)^2 - S^2) # in terms of conservative variables
  # TODO Move this to EoS
  h(ρ, v, ϵ, p) = 1.0 + ϵ + p / ρ            # specific enthalpy
  s(ρ, v, ϵ, p) = log(ϵ / ρ^(γ - 1.0))        # specific entropy
  # primitive variables ρ, v, ϵ, p = P(ρ, ϵ) in terms of conservative ones
  # inverse relations, see Bugner et al. (2018)
  ρ(D, S, τ, p) = D / Wcons(D, S, τ, p)
  v(D, S, τ, p) = S / (τ + D + p)
  ϵ(D, S, τ, p) = (sqrt((τ + p + D)^2 - S^2) - Wcons(D, S, τ, p) * p - D) / D
  # conservative variables D, S, τ in terms of primitive ones
  D(ρ, v, ϵ, p) = ρ * Wprim(ρ, v, p, ϵ)
  S(ρ, v, ϵ, p) = ρ * h(ρ, v, p, ϵ) * Wprim(ρ, v, p, ϵ)^2 * v
  τ(ρ, v, ϵ, p) = ρ * h(ρ, v, p, ϵ) * Wprim(ρ, v, p, ϵ)^2 - p - D(ρ, v, p, ϵ)
  # conservative fluxes
  flux_D(ρ, v, ϵ, p) = D(ρ, v, p, ϵ) * v
  flux_S(ρ, v, ϵ, p) = S(ρ, v, p, ϵ) * v + p
  flux_τ(ρ, v, ϵ, p) = S(ρ, v, p, ϵ) - D(ρ, v, p, ϵ) * v
end


function pressure(eq::Eq, D, S, τ, p0, x)
  @unpack eos, prms = eq
  function residual(p)
    eos_p = EoS.pressure(eos, Formulae.ρ(D, S, τ, p), Formulae.ϵ(D, S, τ, p), x)
    return p - eos_p
  end
  p, _, converged = dg1d.nlsolve(residual, p0, prms)
  if !converged
    error("Could not find pressure `p` that is consistent with equation of state!")
  end
  return p
end


function conservatives(D, S, τ, p)
  ρ = Formulae.ρ(D, S, τ, p) 
  v = Formulae.v(D, S, τ, p) 
  ϵ = Formulae.ϵ(D, S, τ, p) 
  return ρ, v, ϵ
end


@doc """
Maximum characteristic wave speed per point.
TODO Refactor this to max_v (or maybe something else)
TODO Remove @. operations and use Ref in caller to broadcast over arrays
"""
function abs_v(eq::Eq, ρ, v, ϵ, p, x)
  @unpack eos = eq
  cs = EoS.chi(eos, ρ, ϵ, x) + (p / ρ^2) * EoS.kappa(eos, ρ, ϵ, x)
  v0 = v
  vp = (v + cs) / (1 + v * cs)
  vm = (v - cs) / (1 - v * cs)
  return max(map(abs, (v0, vp, vm))...)
end


@doc """
Global maximum absolute characteristic speed.
TODO Refactor this to global_maxabs_v
"""
function maxabs_v(eq::Eq, ρ, v, ϵ, p, x)
  v = abs_v.(Ref(eq), ρ, v, ϵ, p, x)
  return maximum(v[:])
end


@doc """
Pack variables D, S, τ into matrix u.

TODO Think about dispatch Parameters.@unpack macro.
"""
pack(D, S, τ) = vcat(D, S, τ)

@doc """
Unpack variables D, S, τ from matrix u.

TODO Think about dispatch Parameters.@unpack macro.
"""
function unpack(u, x)
  Npts, K = size(x)
  D = view(u, 1:Npts, 1:K)
  S = view(u, Npts+1:2*Npts, 1:K)
  τ = view(u, 2*Npts+1:3*Npts, 1:K)
  return D, S, τ
end
