module AdvectionEq

using dg1d, JLD2, ImageCore

import dg1d: Abstract_BdryConditions, Abstract_ScalarEq, Mesh, AV.Abstract_AV
import Parameters: @with_kw, @unpack

using Plots, PlotThemes

include("types.jl")
include("rhs.jl")
include("initialdata.jl")
include("bdryconditions.jl")

Base.@kwdef struct Eq <: Abstract_ScalarEq
  v::Real = 1.0
end


# implement interface methods
import dg1d: flux, maxabs_v
flux(eq::Eq, u, x) = eq.v * u
maxabs_v(eq::Eq, u, x) = abs(eq.v)
speed(eq::Eq, u, x) = eq.v


struct Eq2 <: Abstract_ScalarEq end
flux(eq::Eq2, u, x) = u.^2
maxabs_v(eq::Eq2, u, x) = 2 * maximum(abs.(u)[:])
speed(eq::Eq2, u, x) = 2 * u


function evolve(;save=false)
  theme(:ggplot2, markersize=1.5)

  N = 10
  K = 80
  CFL = 0.05
  quadr = :LGL_lumped
  scheme = :LSERK4
  tstart = 0.0
  tend = 0.3
  v = 1.0
  type_bdryconditions = :periodic
  type_initialdata = :sine
  xmin = 0.0
  xmax = 1.0
  av_method = :MDA
  av_cmax = 1.0
  av_c1 = 2.5
  av_c2 = 0.2
  filename = "result_burgers.jld"
  plotname = "u_burgers.pdf"

  plot_steps = 15
  plot_dt = (tend - tstart) / plot_steps
  plot_counter = 0

  element = SpectralElement(N, quadr)
  mesh = dg1d.Mesh(element, K, xmin, xmax, CFL)

  u0 = initialdata(type_initialdata, mesh)
  bcs = bdryconditions(type_bdryconditions, u0, mesh)
  # with AV
  av_prms = AV.Parameters(av_method, av_cmax, tstart, av_c1, av_c2)
  av = AV.make_AV(av_prms, mesh)
  # eq = Eq(v)
  eq = Eq2()
  vars = Variables(mesh, av)
  pblm = Problem(vars, mesh, eq, bcs)

  vars.scalar.u .= u0[:] 
  # dg1d.@debug_plot(vars.scalar, mesh)

  save_steps = 40
  save_dt = (tend - tstart) / save_steps
  save_counter = 0

  hist =  [ (tstart, zeros(size(av.μ))) ]

  # interpolation matrix for smoothing operation
  intrp_order = 2
  intrp_element = dg1d.SpectralElement(intrp_order, quadr)
  intrp_from  = element.z
  intrp_to    = intrp_element.z
  intrp_mat   = dg1d.lagrange_interp_matrix(intrp_from, intrp_to)
  # post hook run after each full RK step
  function av_hook(u, t)
    mat_u = reshape(u, (mesh.element.Npts, mesh.K))
    max_v = abs.(flux(eq, mat_u, mesh.x))
    μ = AV.compute(mat_u, max_v, t, pblm.mesh, av)
    av.μ .= dg1d.smooth_C0(μ, intrp_mat)
    # dg1d.@debug_plot(vars.scalar, mesh)
    if (plot_counter * plot_dt < t)
      # dg1d.@debug_plot(vars.visc, mesh)
      # dg1d.@debug_plot(vars.visc, mesh,
      #                  (xlims=(0.4,0.6),))
      # dg1d.@debug_plot(vars.ldgscalar, mesh)
      # dg1d.@debug_plot(vars.scalar, mesh)
      # dg1d.@debug_plot(vars.ldgflux, mesh)
      plot_counter += 1
    end
    if (save_counter * save_dt < t && save)
      push!(hist, (t, deepcopy(av.μ)))
      save_counter += 1
    end
    return true
  end

  # rhs(du, u, t) = rhs!(du, u, t, pblm, formulation=:weak)
  rhs(du, u, t) = rhs!(du, u, t, pblm, formulation=:weak)

  dt = AV.timestep(u0, mesh, eq, av)
  tspan = (tstart, tend)
  u_num = dg1d.evolve_ERK(rhs, u0, dt, tspan, scheme, post_hook=av_hook)

  u_ana = @. sin(π * (mesh.x - (tend - tstart)))

  output_dir = dg1d.get_output_dir()
  imgname = "2019_yu_av_burgers_p7.png"
  imgname = joinpath(output_dir, imgname)
  img = load(imgname)

  plt = plot()
  dg1d.plot(vars.scalar, mesh; plts=[plt])


  if save
    output_dir = dg1d.get_output_dir()
    filename = joinpath(output_dir, filename)
    jldsave(filename; hist, x=mesh.x, Npts=element.Npts, K=mesh.K)
    plotname = joinpath(output_dir, plotname)

    plt = plot()
    plot!(plt, [0.4,0.6], [-0.8,0.8], reverse(img, dims=1),
          yflip=false, aspect_ratio=:none)
    dg1d.plot(vars.scalar, mesh; xlims=(0.4,0.6), ylims=(-0.8,0.8), plts=[plt])
    savefig(plt, plotname)
  end

end


function plot_av_evolution(;save=false)

  output_dir = dg1d.get_output_dir()
  filename = "result_burgers.jld"
  filename = joinpath(output_dir, filename)
  plotname = "av_burgers.pdf"
  plotname = joinpath(output_dir, plotname)

  x = nothing
  hist = nothing
  Npts = nothing
  K = nothing
  jldopen(filename, "r") do file
    x = read(file, "x")
    hist = read(file, "hist")
    Npts = read(file, "Npts")
    K = read(file, "K")
  end

  xs = x[:]
  ts = [ tpl[1] for tpl in hist ]
  μs = zeros(Float64, (length(ts), length(xs)))
  for (i, tpl) in enumerate(hist)
    μs[i,:] = tpl[2][:]
  end

  # partition xs and μs at domain boundaries, needed for contour plot
  part_xs = [ xs[1+(k-1)*Npts:k*Npts] for k in 1:K ]
  part_μs = [ μs[:,1+(k-1)*Npts:k*Npts] for k in 1:K ]

  theme(:ggplot2)
  Plots.gr()
  plt = contour()
  for (pxs, pμs) in zip(part_xs, part_μs)
    contour!(plt, pxs, ts, pμs, fill=true, linewidth=0)
  end
  plot!(plt, xlims=(0.4,0.6), colorbar_title="Artificial viscosity")
  display(plt)
  if save
    savefig(plt, plotname)
  end

end


end # module
