module HeatEq

using dg1d, LinearAlgebra
import dg1d: Abstract_BdryConditions, Abstract_ScalarEq, Mesh, AV.Abstract_AV
import Parameters: @with_kw, @unpack

using Plots, PlotThemes

include("types.jl")
include("rhs.jl")

struct Eq <: Abstract_ScalarEq end

# implement interface methods
import dg1d: flux, maxabs_v
flux(eq::Eq, u, x) = 0.0
maxabs_v(eq::Eq, u, x) = 0.0


function evolve()
  theme(:ggplot2, #linewidth=0.5, markerstrokewidth=0.1,
        markersize=1.5)

  N = 5
  K = 80
  CFL = 0.05
  quadr = :GLGL
  scheme = :LSERK4
  type_bdryconditions = :periodic
  type_initialdata = :sine
  tstart = 0.0
  tend = 0.3 # 30.3
  xmin = 0.0
  xmax = 2.0 * π
  av_method = :MDA
  av_cmax = 10.0
  av_c1 = 2.5
  av_c2 = 0.2

  plot_steps = 15
  plot_dt = (tend - tstart) / plot_steps
  plot_counter = 0

  element = SpectralElement(N, quadr)
  mesh = dg1d.Mesh(element, K, xmin, xmax, CFL)
  # display(element.invM)
  u0 = dg1d.AdvectionEq.initialdata(type_initialdata, mesh)
  bcs = dg1d.AdvectionEq.bdryconditions(type_bdryconditions, u0, mesh)
  eq = Eq()
  av_prms = AV.Parameters(av_method, av_cmax, tstart, av_c1, av_c2)
  av = AV.make_AV(av_prms, mesh)
  vars = Variables(mesh, av)
  pblm = Problem(vars, mesh, eq, bcs)

  vars.scalar.u .= u0[:]
  vars.visc.μ .= 1.0
  # dg1d.@debug_plot(vars.scalar, mesh)


  function hook(u, t)
    mat_u = reshape(u, (mesh.element.Npts, mesh.K))
    if (plot_counter * plot_dt < t)
      # display(μ)
      # display(t)
      # dg1d.@debug_plot(vars.visc, mesh)
      # dg1d.@debug_plot(vars.scalar, mesh)
      # dg1d.@debug_plot(vars.ldgscalar, mesh)
      # dg1d.@debug_plot(vars.ldgflux, mesh)
      plot_counter += 1
    end
    return true
  end

  rhs(du, u, t) = rhs!(du, u, t, pblm)

  mindx = minimum(abs.((diff(mesh.x, dims=1))))
  dt = CFL * (mindx)^2

  display(dt)
  tspan = (tstart, tend)
  u_num = dg1d.evolve_ERK(rhs, u0, dt, tspan, scheme, post_hook=hook)
  u_ana = @. exp(-(tend-tstart)) * sin(mesh.x)

  vars.scalar.u .= u_num[:]
  dg1d.plot(vars.scalar, mesh)

  display(norm(u_num - u_ana))

  dg1d.@debug_plot(vars.scalar, mesh)

  return
end


end # module
