export ScalarEq
module ScalarEq


using dg1d
import dg1d: Abstract_ScalarEq, Abstract_BdryConditions, AV.Abstract_AV, Mesh
import Parameters: @unpack, @with_kw


include("initialdata.jl")
include("bdryconditions.jl")


function timestep(u, mesh::Mesh, eq::Abstract_ScalarEq)
  @unpack dl, x, CFL, element = mesh
  @unpack N = element
  vmax = maxabs_v(eq, u, x)
  dt = CFL * dl / (N * vmax)
  return dt
end


function rhs!(du, u, t, mesh::Mesh, eq::Abstract_ScalarEq, bc::Abstract_BdryConditions, ::Nothing)
  @unpack dl, K, x, element = mesh
  @unpack N = element

  isperiodic = dg1d.isperiodic(bc)
  if !isperiodic
    nu_lhs, nu_rhs = dg1d.bdry_flux(u, t, bc)
    u[1], u[end] = nu_lhs, nu_rhs
  end

  f = flux(eq, u, x)
  intf_f  = dg1d.to_interface(f, mesh)
  intf_u  = dg1d.to_interface(u, mesh)
  max_v = maxabs_v(eq, u, x)
  nf_lhs, nf_rhs = dg1d.lax_friedrich_flux(intf_f, intf_u, max_v)
  if !isperiodic
    nu_lhs, nu_rhs  = dg1d.bdry_flux(u, t, bdryconds)
    nf_lhs[1]       = flux(eq, nu_lhs, x[1])
    nf_rhs[end]     = flux(eq, nu_rhs, x[end])
  end

  dg1d.rhs_weak!(du, f, nf_lhs, nf_rhs, element, mesh)
  return
end


function rhs!(du, u, t, mesh::Mesh, eq::Abstract_ScalarEq, bc::Abstract_BdryConditions,
    av::Abstract_AV)
  @unpack dl, K, x, element = mesh
  @unpack N = element
  @unpack μ = av

  isperiodic = dg1d.isperiodic(bc)
  if !isperiodic
    nu_lhs, nu_rhs = dg1d.bdry_flux(u, t, bc)
    u[1], u[end] = nu_lhs, nu_rhs
  end

  ## auxiliary equation: q - ∂x u = 0
  intf_u = dg1d.to_interface(u, mesh)
  nu_lhs, nu_rhs = dg1d.central_flux(intf_u)
  if !isperiodic
    nu_lhs[1], nu_rhs[end] = dg1d.bdry_flux(u, t, bdryconds)
  end
  q = similar(u)
  dg1d.rhs_weak!(q, -u, -nu_lhs, -nu_rhs, mesh)

  ## reduced equation: ∂t u + ∂x f - ∂x μ q = 0
  f = dg1d.flux(eq, u, x)
  g = - μ .* q
  intf_f  = dg1d.to_interface(f, mesh)
  intf_u  = dg1d.to_interface(u, mesh)
  intf_g  = dg1d.to_interface(g, mesh)
  # bdry values relevant for max_v?
  max_v = dg1d.maxabs_v(eq, u , x)
  nf_lhs, nf_rhs = dg1d.lax_friedrich_flux(intf_f, intf_u, max_v)
  ng_lhs, ng_rhs = dg1d.central_flux(intf_g)
  if !isperiodic
    u_lhs, u_rhs = dg1d.bdry_flux(u, t, bdryconds)
    x_lhs, x_rhs = x[1], x[end]
    nf_lhs[1]   = flux(eq, u_lhs, x_lhs)
    nf_rhs[end] = flux(eq, u_rhs, x_rhs)
    ng_lhs[1]   = 0.0
    ng_rhs[end] = 0.0
  end

  nfg_lhs = nf_lhs + ng_lhs
  nfg_rhs = nf_rhs + ng_rhs
  fg = f + g
  dg1d.rhs_weak!(du, fg, nfg_lhs, nfg_rhs, mesh)

  return
end


struct PostHookAV
  hook::Function
  function PostHookAV(mesh::Mesh, eq::Abstract_ScalarEq, av::Abstract_AV)
    # @unpack intrp_order = av
    @unpack element, x = mesh
    @unpack quadr, z = element

    # interpolation matrix for smoothing operation
    # intrp_order = smooth_order
    intrp_order = 1
    intrp_element = dg1d.SpectralElement(intrp_order, quadr)
    intrp_from  = element.z
    intrp_to    = intrp_element.z
    intrp_mat   = dg1d.lagrange_interp_matrix(intrp_from, intrp_to)

    # post hook run after each full RK step
    function hook(u, t)
      max_v = dg1d.flux(eq, u, x)
      μ = AV.compute(u, max_v, t, mesh, eq, av)
      av.μ .= dg1d.smooth_C0(μ, intrp_mat)
      return true
    end

    return new(hook)
  end
end






# function plot(u::AbstractMatrix, pblm::Problem, av::Union{AV.AbstractAV,Nothing}=nothing)
#
#   @unpack prms, mesh = pblm
#   @unpack K = prms
#   @unpack x = mesh
#
#   Plots.gr()
#
#   # Temporary fix to cure problem with InexactError occuring
#   # due to too small values in u
#   # Might be either a bug in Plots package or a problem with my installation
#   # It is a bug with my installation: https://github.com/void-linux/void-packages/pull/30061
#   # and https://github.com/JuliaLang/julia/issues/40198
#   dg1d.purge_zeros!(u)
#
#   nplts = 1
#   if !isnothing(av)
#     nplts += 1
#   end
#
#   xlabel = "x"
#   plts = [ Plots.plot() for _ in 1:nplts ]
#   vars = [ u ]
#   ylabels = [ "u(t,x)" ]
#   if !isnothing(av)
#     push!(vars, av.μ)
#     push!(ylabels, "μ(t,x)")
#   end
#
#   for (plt, var, ylabel) in zip(plts, vars, ylabels)
#     for k in 1:K
#       Plots.plot!(plt, x[:,k], var[:,k], label=false)
#     end
#     Plots.plot!(plt, ylabel=ylabel)
#   end
#
#   plt = Plots.plot(plts..., layout=(nplts,1), size=(1200,1200))
#   Plots.gui(plt)
#
#   return
# end
#
#
# @doc """
# TODO Agree on one time step computation.
# TODO Implement adaptive time stepping computation.
# """



# evolve_problem(u0, pblm::Problem, ::Nothing) = evolve_problem(u0, pblm)
#
#
# function evolve_problem(u0, pblm::Problem, av::AV.AbstractAV)
#
#   @unpack prms, element, mesh, Eq = pblm
#   @unpack L, tend = prms
#   @unpack x = mesh
#
#   dt = AV.timestep(u0, pblm, av)
#
#   # interpolation matrix for smoothing operation
#   intrp_order = 1
#   intrp_element = dg1d.SpectralElement(intrp_order, prms.quadr)
#   intrp_from  = pblm.element.z
#   intrp_to    = intrp_element.z
#   intrp_mat   = dg1d.lagrange_interp_matrix(intrp_from, intrp_to)
#
#   # post hook run after each full RK step
#   function hook(u, t)
#     max_v = Eq.df.(u, x)
#     μ = AV.compute(u, max_v, t, pblm, av)
#     av.μ .= dg1d.smooth_C0(μ, intrp_mat)
#     return true
#   end
#
#   # ODE problem
#   F!(du, u, t) = AV_rhs!(du, u, t, pblm, av)
#   tspan = (prms.tstart, prms.tend)
#   u_num = dg1d.evolve_ERK(F!, u0, dt, tspan, scheme=prms.scheme, post_hook=hook)
#
#   return u_num
# end
#

end # module
