using dg1d



@with_kw struct Problem
  mesh::dg1d.Mesh
  element::dg1d.SpectralElement
  prms::Parameters
end


function Problem(prms::Parameters)
  @unpack N, K, L, quadr = prms
  element   = dg1d.SpectralElement(N, quadr)
  mesh      = dg1d.Mesh(K, L, element.z)
  return Problem(mesh, element, prms)
end



function rhs!(du, u, t, problem::Problem)
  @unpack element, mesh, prms = problem
  @unpack K, x = mesh
  @unpack N = element
  @unpack isperiodic = prms

  ## auxiliary equation: 0 = q - ∂x u =: q + ∂x v
  v = - u
  intf_v = dg1d.to_interface(v, element, mesh)
  nv_lhs, nv_rhs = dg1d.central_flux(intf_v)
  q = similar(v)
  dg1d.rhs_weak!(q, v, nv_lhs, nv_rhs, element, mesh)

  ## reduced equation: 0 = ∂t u - ∂x q =: ∂t u + ∂x g
  g = - q
  intf_g  = dg1d.to_interface(g, element, mesh)
  ng_lhs, ng_rhs = dg1d.central_flux(intf_g)
  dg1d.rhs_weak!(du, g, ng_lhs, ng_rhs, element, mesh)

  return
end # function


function timestep(u, problem::Problem)
  @unpack element, mesh, prms = problem
  @unpack x = mesh
  @unpack CFL = prms
  @unpack N = element

  mindx = minimum(abs.((diff(x, dims=1))))
  dt = CFL * (mindx)^2

  return dt
end # function



function plot(u, u_ana, pblm::Problem)

  @unpack prms, mesh = pblm
  @unpack quadr, N, K, scheme = prms
  cfg = """set title "Heat equation, $quadr quadrature, """ *
  """N = $N, K = $K, scheme = $scheme";""" *
  """set xlabel "x"; set ylabel "u(t,x)";"""
  dg1d.dg_plot([u, u_ana], mesh, keys=["numerical", "analytical"], 
    plotcfg=cfg, session=:heat_solution)

  err = abs.(u - u_ana)

  cfg = """set title "Heat equation, $quadr quadrature, """ *
  """N = $N, K = $K, scheme = $scheme";""" *
  """set xlabel "x"; set ylabel "ϵ(x)";"""
  dg1d.dg_plot([err], mesh, keys=["error"], 
    plotcfg=cfg, session=:heat_error)

end


function evolve(;N::Int = 10, K::Int = 3, CFL = 0.2,
                show::Bool = true)

  prms = Parameters(
    N = N,
    K = K,
    L = 2.0 * π,
    CFL = CFL,
    quadr = :LGL,
    scheme = :SSPRK3
  )
  display(prms)
  pblm = Problem(prms)

  @unpack mesh = pblm
  @unpack L, x = mesh
  @unpack tend, tstart = prms
  u0 = @. sin(x)

  function hook(u, t)
    @unpack prms, mesh = pblm
    dg1d.dg_plot([u], mesh, keys=["numerical"], session=:Heat)
    return true
  end

  dt = timestep(u0, pblm)
  F!(du, u, t) = rhs!(du, u, t, pblm)
  tspan = (tstart, tend)

  u_num = deepcopy(u0)
  try
    u_num .= dg1d.evolve_ERK(F!, u0, dt, tspan, scheme=prms.scheme)
  catch e
    if isa(e,InterruptException)
      println("\nUser terminated ...")
      return
    else
      rethrow(e)
    end
  end
  
  u_ana = @. exp(-(tend - tstart)) * sin(x)

  if show
    plot(u_num, u_ana, pblm)
  end

  return
end


# function solve(;N::Int = 10, K::Int=3, CFL = 0.20)
#
#   β           = 0.0
#   τ           = 0.0
#   scheme      = :SSPRK3
#   L           = 2.0 * π
#   quadr       = :LGL
#
#   element     = dg1d.SpectralElement(N, quadr)
#   mesh        = dg1d.Mesh(K, L, element.z, grid_offset = -L/2.0)
#   u0          = initialdata(:sine, mesh)
#   isperiodic  = true
#   problem     = Problem(mesh, element, CFL, isperiodic)
#
#   # time interval
#   tspan = (0.0, 1.8)
#   dt = timestep(u0, problem)
#
#   # ODE problem
#   F!(du, u, t) = rhs!(du, u, t, problem, β, τ)
#   u_sol = dg1d.evolve_ERK(F!, u0, dt, tspan, scheme=scheme)
#
#   # error estimation
#   u_ana = u_analytical.(mesh.x, tspan[2])
#
#   cfg = """ set title "Heat equation, $(quadr) quadrature, """ *
#   """N = $N, K = $K, scheme = $scheme"; set xlabel "x"; set ylabel "u(t,x)"; """
#   dg1d.dg_plot([u_sol, u_ana], mesh, keys=["u", "u_{ana}"], plotcfg=cfg)
#
# end # function


# function test_cvgrate(;CFL=0.25, β=0.0, τ=0.0)
#   s(;N, K, CFL) = solve(N=N, K=K, CFL=CFL, cvgtest=true, β=β, τ=τ)
#   N, K, err = Analysis.eval_cvgrate(s, CFL, maxN = 4, maxK=80)
#   title = "Heat equation, periodic domain, CFL = $CFL, β = $β, τ = $τ"
#   Analysis.plot_cvgrate(N, K, err, title)
# end # function


#######################################################################
#                        FOR FUTURE REFERENCE                         #
#######################################################################


# function impose_bc_v(intf_v)
#   # intf_v.lhs[1] = intf_v.rhs[1]
#   # intf_v.lhs[1] = intf_v.rhs[1] / 3.0
#   # intf_v.rhs[end] = 3.0 / 2.0 * intf_v.lhs[end]
#   # intf_v.lhs[1] = - intf_v.rhs[1]
#   # intf_v.rhs[end] = - intf_v.lhs[end]
#   # intf_v.lhs[1] = 0.0
#   # intf_v.rhs[end] = 0.0
# end # function
#
#
#
# function impose_bc_g(intf_g)
#   # intf_g.lhs[end] = intf_g.rhs[end]
#   # intf_g.lhs[1] = intf_g.rhs[1]
#   # intf_g.rhs[end] = intf_g.lhs[end]
#   # intf_g.lhs[1] = intf_g.rhs[1]
#   # intf_g.rhs[end] = intf_g.lhs[end]
#   # intf_g.lhs[1] = exp(-t)
#   # intf_g.rhs[end] = exp(-t)
# end # function
