@doc """
  advection_rhs!(du, u, t, pblm::Problem, [formulation])

Compute RHS expression of advection-diffusion equation

  ∂t u = - ∂x f(u) + ∂x μ(u) ∂x u = - ∂x f(u) - ∂x μ(u) g(u)

using a DG scheme in strong form with central fluxes and return result inplace
in `du`.

Supported `formulation`s are `:strong` and `:weak`.
"""
function advection_rhs!(du, u, t, pblm::Problem; formulation=:strong)
  @unpack eq, mesh, vars, bcs = pblm
  @unpack dl, K, x, element = mesh
  @unpack Npts = element

  shape = (Npts, K)
  vars.scalar.u .= u[:]

  mat_u = reshape(vars.scalar.u, shape)
  mat_f = reshape(vars.flux.f, shape)
  mat_μ = reshape(vars.visc.μ, shape)
  @. mat_f = flux(eq, mat_u, x)

  intf_f  = dg1d.to_interface(mat_f, mesh)
  intf_u  = dg1d.to_interface(mat_u, mesh)

  # max_v = maxabs_v(eq, mat_u, x)
  # cell local wave speed
  v = speed.(eq, mat_u, mesh.x)
  max_v = dg1d.local_maxabs(v, mesh)
  # max_v = maximum(abs.(mat_f), dims=1)[:]
  push!(max_v, max_v[end])
  max_v = @. 0.5 * (max_v + $circshift(max_v, 1))
  nf_lhs, nf_rhs = dg1d.lax_friedrich_flux(intf_f, intf_u, max_v)
  # nf_lhs, nf_rhs = dg1d.central_flux(intf_f)
  nfi_lhs, nfi_rhs = dg1d.inner_flux(intf_f)

  mat_du = reshape(du, shape)
  dg1d.rhs_strong!(mat_du, mat_f, nfi_lhs - nf_lhs, nfi_rhs - nf_rhs, mesh)

  return
end



@doc """
  ldg_aux_rhs!(q, u, pblm::Problem)

Compute RHS expression of auxiliary LDG equation

  q = - ∂x u

using a DG scheme in strong form with central flux.
Result is returned inplace in `q`.

Supported `formulation`s are `:strong` and `:weak`.
"""
function ldg_aux_rhs!(q, u, pblm::Problem; formulation=:strong)

  @unpack eq, mesh = pblm
  @unpack K, x, element = mesh
  @unpack Npts = element

  shape = (Npts, K)
  mat_q = reshape(q, shape)
  mat_u = reshape(u, shape)

  ## auxiliary equation: q + ∂x u = 0
  intf_u = dg1d.to_interface(mat_u, mesh)
  nu_lhs, nu_rhs = dg1d.central_flux(intf_u)

  if formulation === :strong
    nui_lhs, nui_rhs = dg1d.inner_flux(intf_u)
    dg1d.rhs_strong!(mat_q, mat_u, nui_lhs-nu_lhs, nui_rhs-nu_rhs, mesh)
  elseif formulation === :weak
    dg1d.rhs_weak!(mat_q, mat_u, nu_lhs, nu_rhs, mesh)
  else
    error("Invalid DG formulation '$formulation' requested!")
  end


  return
end



@doc """
  rhs!(du, u, t, pblm::Problem, [formulation])

Compute RHS expression of advection-diffusion equation

  ∂t u = - ∂x f(u) + ∂x μ(u) ∂x u = - ∂x f(u) - ∂x μ(u) g(u)

using a DG scheme in strong form with central fluxes.
Result is returned inplace in `du`.

Supported `formulation`s are `:strong` and `:weak`.
"""
function rhs!(du, u, t, pblm::Problem; formulation=:strong)
  @unpack eq, mesh, vars, bcs = pblm
  @unpack dl, K, x, element = mesh
  @unpack Npts = element

  shape = (Npts, K)
  vars.scalar.u .= u[:]

  ## auxiliary equation: q + ∂x u = 0
  ldg_aux_rhs!(vars.ldgscalar.q, vars.scalar.u, pblm, formulation=formulation)

  mat_u = reshape(vars.scalar.u, shape)
  mat_f = reshape(vars.flux.f, shape)
  mat_μ = reshape(vars.visc.μ, shape)
  mat_q = reshape(vars.ldgscalar.q, shape)
  mat_g = reshape(vars.ldgflux.g, shape)
  @. mat_f = flux(eq, mat_u, x)
  @. mat_g = mat_μ * mat_q

  intf_f  = dg1d.to_interface(mat_f, mesh)
  intf_u  = dg1d.to_interface(mat_u, mesh)
  intf_g  = dg1d.to_interface(mat_g, mesh)

  v = speed.(eq, mat_u, mesh.x)
  max_v = dg1d.local_maxabs(v, mesh)
  # max_v = dg1d.local_maxabs(v)
  # max_v = maximum(abs.(mat_f), dims=1)[:]
  push!(max_v, max_v[end])
  max_v = @. 0.5 * (max_v + $circshift(max_v, 1))
  nf_lhs, nf_rhs = dg1d.lax_friedrich_flux(intf_f, intf_u, max_v)
  # nf_lhs, nf_rhs = dg1d.central_flux(intf_f)
  ng_lhs, ng_rhs = dg1d.central_flux(intf_g)

  ## advection-diffusion equation: ∂t u + ∂x (f + g) = 0
  mat_du = reshape(du, shape)
  if formulation === :strong
    nfi_lhs, nfi_rhs = dg1d.inner_flux(intf_f)
    ngi_lhs, ngi_rhs = dg1d.inner_flux(intf_g)
    dg1d.rhs_strong!(mat_du, mat_f+mat_g, 
                     (nfi_lhs+ngi_lhs) - (nf_lhs+ng_lhs), 
                     (nfi_rhs+ngi_rhs) - (nf_rhs+ng_rhs),
                     mesh)
  elseif formulation === :weak
    dg1d.rhs_weak!(mat_du, mat_f+mat_g, nf_lhs+ng_lhs, nf_rhs+ng_rhs, mesh)
  else
    error("Invalid DG formulation '$formulation' requested!")
  end


  return
end
