# TODO 
#   - Think about substituting Vector{DataType} by NamedTuple so
#   that one can access components by familiar names.
#   - Replace components::Vector{DataType} with Tuple{DataType}
#   - Figure out which constructors are needed
#     - Make StateVector from given data and components,
#       data already allocated before.
#   - Vector{DataType} is not type stable! Take care of this!
#     Maybe something along the lines of this suggestion
#     https://discourse.julialang.org/t/parametric-type-in-an-variable-length-tuple/9109
#     together with dispatching on symbols with the Val type?
#     But I think this will still be instable in the end...



export StateVector, get_component

@doc """
Wrapper around an array `data` which bundles values data points of the list
of components in consecutive order, e.g.
  data = [ rho_1, rho_2, ..., rho_N, p_1, p_2, ... p_N, etc. ]

Note: StateVector is a wrapper around an already allocated array and does not
allocate by itself.
"""
struct StateVector{T}
  data::Vector{T}
  components::Vector{DataType}
  # components::Vector{DataType}
  function StateVector(data::Vector{T}, components::Vector{DataType}) where T 
    for component in components
      @assert component <: AbstractVariables "Component '$component' must be AbstractVariables!"
    end
    nvars_total = sum([ n_variables(comp) for comp in components ])
    @assert length(data) % nvars_total == 0 "Mismatch in number of data points and components!"
    return new{T}( data, components )
  end
end

# Remove this, because we don't want StateVector to make new allocations.
function StateVector(::Type{T}, Npts::Int, components::Vector{DataType}) where T
  Nvars = sum([ n_variables(comp) for comp in components ])
  data = zeros(T, Npts * Nvars)
  return StateVector(data, components)
end

Base.length(sv::StateVector) = length(sv.data)
n_components(sv::StateVector) = length(sv.components)
n_variables(sv::StateVector) = sum([ n_variables(component) for component in sv.components])


@doc """
Get `idx`th component of data vector.
"""
function get_component(idx::Int, sv::StateVector)
  n = length(sv)
  nvars = n_variables(sv)
  npts = Int(n / nvars)
  idx_start = if idx == 1
    1
  else
    npts * sum([ n_variables(component) for component in sv.components[1:idx-1] ])
  end
  idx_length = npts * n_variables(sv.components[idx])
  component = sv.components[idx]
  # return component(view(sv.data, idx_start:idx_start+idx_length-1))
  return to_variable(component, view(sv.data, idx_start:idx_start+idx_length-1))
end


# TODO Implement get_component but with symbol instead of index to access
# components from named tuple.

Base.vec(sv::StateVector) = sv.data
