@doc """
Scan matrix M for elements small than thrs and set them to 0.0.
"""
function purge_zeros!(M, thrs::Float64=1e-13)
  n = length(M)
  M .= [ (abs(m) < thrs) ? 0.0 : m for m in M ]
  return
end


function get_outputdir()
  thisdir = dirname(@__FILE__)
  outputdir = "$thisdir/../outputs/"
  outputdir = normpath(outputdir)
  if !isdir(outputdir)
    mkdir(outputdir)
  end
  return outputdir
end


function get_output_dir()
  return get_outputdir()
end

@deprecate get_output_dir get_outputdir


@doc """
Least squares fit of linear model function.
Alogrithm taken from https://mathworld.wolfram.com/LeastSquaresFitting.html
x, y  ... Vectors containing data points
f     ... linear fit to data points
a, b  ... coefficients of linear model: f(x) = a + b * x
"""
function least_sqr_linear(x::tV, y::tV) where tV <: Vector{Float64}

  n = length(x)
  @assert n == length(y) "Mismatch in number of data points x, y!"

  yx = dot(y, x)
  xx = dot(x, x)
  mean_y = sum(y) / n
  mean_x = sum(x) / n

  ss_xx = xx - mean_x^2 * n
  ss_xy = yx - mean_x * mean_y * n

  b = ss_xy / ss_xx
  a = mean_y - b * mean_x

  f(x) = a + b * x
  return f, a, b
end


@with_kw struct NlSolveParameters
  abstol::Float64 = 1e-10
  reltol::Float64 = 1e-10
  step::Float64   = 1e-8
  maxits::Int64   = 10
end


const Default_NlSolveParameters = NlSolveParameters()


# TODO Rename this to nlsolve_fd_approx
function nlsolve(f::Function, x0::Float64; prms::NlSolveParameters=NlSolveParameters())
  @unpack_NlSolveParameters prms
  xi = x0
  it = 1
  fi = f(xi)
  tol = reltol * abs(fi) + abstol
  converged = false
  while true
    x0 = xi
    fi = f(x0)

    # scaling FD perturbation
    # C.T.Kelly - Solving Nonlinear Equations with Newton's Method, 2003
    # Section 2.3
    if isapprox(x0, 0.0)
      dx = step
      dfi = (f(dx) - fi) / dx
    else
      dx = step * max(abs(x0), 1.0) * sign(x0)
      dfi = (f(x0+dx) - fi) / dx
    end # if

    # update iterate
    dfi = (f(x0+dx) - fi) / dx
    xi = x0 - fi / dfi

    if abs(fi) <= tol
      converged = true
      break
    end

    it += 1
    if isnan(xi) || isinf(xi) || it > maxits
      break
    end
  end # while
  return xi, fi, converged
end



# TODO Rename this to nlsolve
function nlsolve_exact(f_df::Function, x0::Float64;
    prms::NlSolveParameters=NlSolveParameters())
  @unpack_NlSolveParameters prms
  xi = x0
  fi, dfi = f_df(xi)
  tol = reltol * abs(fi) + abstol
  # check if guess is already good enough
  # if abs(fi) <= tol
  #   return x0, true
  # end
  converged = false
  it = 1
  while !converged && it <= maxits
    # update iterate
    xi = x0 - fi / dfi
    # sanity check
    if isnan(xi) || isinf(xi)
      break
    end
    fi, dfi = f_df(xi)
    # test iterate
    if abs(fi) <= tol
      converged = true
    end
    x0 = xi
    it += 1
  end # while
  return xi, fi, converged
end


stepdata(x, x0, fL, fR) = [ (xi < x0) ? fL : fR for xi in x ]


# stolen from https://discourse.julialang.org/t/convert-type-to-dictionary/6982/7
typedict(x::T) where T = Dict(fn=>getfield(x,fn) for fn ∈ fieldnames(T))


@doc """
Finite difference approximation of first and second order derivatives.
Result is computed on grid points r[2:end-1], e.g. no derivative
is computed at domain endpoints

TODO Add derivatives at endpoints with one-sided FD formulae.
"""
function fd_approx(f, r)
  dr = diff(r)
  fr = zeros(length(f)-2)
  frr = zeros(length(f)-2)
  for i in 1:length(fr)
    dm1 = dr[i]
    dp1 = dr[i+1]
    dpd = dm1 + dp1
    fr[i] = - dp1 / (dm1*dpd) * f[i] + (dp1 - dm1) / (dm1*dp1) * f[i+1] + dm1 / (dp1 * dpd) * f[i+2]
    frr[i] = 2.0 / (dm1*dpd) * f[i] - 2.0 / (dm1*dp1) * f[i+1] + 2.0 / (dp1 * dpd) * f[i+2]
  end
  return fr, frr
end


#######################################################################
#              Wrappers to conviniently include scripts               #
#######################################################################

# Requires DrWatson as dependency, but this is not really what I want, since
# then using dg1d will always include those files again

# function include_script_fns()
#   jlfiles = filter(x->occursin(".jl",x), readdir(scriptsdir()))
#
#   e = quote end
#   for file in jlfiles
#     fname = "script_" * file
#     fname = replace(fname,".jl"=>"")
#     fname = replace(fname," "=>"_")
#     fname = Symbol(fname)
#     blk = quote
#       export $(fname)
#       $(fname)() = include($(scriptsdir(file)))
#     end
#     append!(e.args,blk.args)
#   end
#   eval(e)
#   return
# end
#
# include_script_fns()
