## TODO
# - Think about how one would handle tensors this way?
#   Specifically, how to store components and acount for symmetries?
#   Options:
#     - Use wrapper around variable that holds infos about symmetries etc.
#       +: Does not alter current implementation.
#       -: Might require to decouple variables, e.g. split conservatives
#       into scalar and vector quantities ...
#     - Make AbstractVariables a parameteric type and dispatch on it ...
#       Does not makes sense, because with the VariableStruct macro we already
#       generate different subtypes of AbstractVariables on which we already
#       dispatch. Also: A subtype of AbstractVariables should always have the
#       same 'symmetries'.
#     - Can this even work out? Or am I trying to combine to different concepts,
#       e.g. bundling variables with different symmetries but treating them 
#       equally? This does not sound right ...
#     - Back to the wrapper idea: Make it so that all operations are passed
#       forward to the data as if it were not even there. Cf. NewEoS variables.
#       Open Question: Type stability?
#       Related discussion:
#       https://discourse.julialang.org/t/traitwrapper-an-idea-for-developing-a-traits-based-system/31179
#       Packages:
#       https://github.com/xiaodaigh/TraitWrappers.jl
#       https://github.com/mauro3/SimpleTraits.jl
#
# - Think about naming convention for vector/tensor types, e.g.
#   How to label
#     - components
#     - scalar derivatives of components
#   consistently.
#   Ideas:
#     - Use lower and upper case letters to distinguish lower and upper case
#     components.
#       +: Consistent
#       -: Lower and upper cased non-alphabetic characters can be hard to
#         distinguish depending on the used font.
#         Also: Do I really want to continue using unicode symbols in the code?
#         Or should I start dropping them for the sake of keeping the code 
#         more 'accesible' ...?
#     - Use underscores to append derivative dependency.
#       +: Makes sense. Also: Derivatives are almost always lower cased, or are they?
#     - Use symbol to indicate upper/lower 'casedness'?
#       Makes it explicit but also verbose. Q: Which do I prefer?
#     - Only ever consider lower/upper cased variables?
#       +: Consistency.
#       -: Requires to rewrite all formulae.
#     - Remove derivative syntax and force to 'alias' any differentiated
#       quantities with a new variable name, e.g. Γxyz := ∂z g_xy.
#       +: Consistent.
#       -: Can't think of a drawback right now ...

export AbstractVariables, n_variables, to_variable, fields, @VariableStruct

abstract type AbstractVariables end
export n_variables

function Base.isnan(var::AbstractVariables)
  varnans = [ any(isnan.(getfield(var, f))) for f in fieldnames(typeof(var)) ]
  return any(varnans)
end

function Base.isinf(var::AbstractVariables)
  varnans = [ any(isinf.(getfield(var, f))) for f in fieldnames(typeof(var)) ]
  return any(varnans)
end


n_variables(var::AbstractVariables) = n_variables(typeof(var))
n_variables(::Type{T}) where T<:AbstractVariables = fieldcount(T)


function (::Type{TVar})(Npts::Int, ::Type{T}=Float64) where {TVar <: AbstractVariables, T}
  zs = [ (f, zeros(T, Npts)) for f in fieldnames(TVar) ]
  return TVar(; zs...)
end


function to_variable(::Type{TVar}, data::AbstractArray{T,1}) where {TVar <: AbstractVariables, T}
  N = n_variables(TVar)
  @assert length(data) % N == 0
  Npts = Int(length(data) / N)
  var_views = [ (f, view(data, (i-1)*Npts+1:i*Npts)) for (i,f) in enumerate(fieldnames(TVar)) ]
  return TVar(; var_views...)
end


function Base.vec(var::AbstractVariables)
  return cat([getfield(var,name) for name in fieldnames(typeof(var))]..., dims=1)
end


function fields(var::AbstractVariables)
  return [ getfield(var, f) for f in fieldnames(typeof(var)) ]
end


macro VariableStruct(struct_name, var_names...)
  esc(
    quote

      struct $struct_name <: dg1d.AbstractVariables
        $(var_names...)
        $struct_name(;$(var_names...)) = new($(var_names...))
      end

    end
  )
end


function Base.show(io::IO, var::AbstractVariables)
  data = hcat([f for f in fields(var)]...)
  header = [ fn for fn in fieldnames(typeof(var)) ]
  println()
  pretty_table(data, header=header)
end


struct Segment{T<:AbstractVariables}
  var::T
  mesh::Mesh
  pts::UnitRange
  cells::UnitRange
end


Segment(var, mesh, i::Int, cells::UnitRange) = Segment(var, mesh, i:i, cells)
Segment(var, mesh, pts::UnitRange, j::Int) = Segment(var, mesh, pts, j:j)
Segment(var, mesh, i::Int, j::Int) = Segment(var, mesh, i:i, j:j)
Segment(var, mesh, ::Colon, cells) = Segment(var, mesh, 1:mesh.element.Npts, cells)
Segment(var, mesh, pts, ::Colon) = Segment(var, mesh, pts, 1:mesh.K)


const cell_colors = [ :red, :green, :yellow, :blue, :magenta, :cyan ]

@inline periodic_index(i, N) = (i % N == 0) ? N : i % N

function Base.show(io::IO, seg::Segment{T}) where T
  @unpack var, mesh, pts, cells = seg
  shape = (mesh.element.Npts, mesh.K)
  data_x = mesh.x[pts,cells][:]
  data_f = hcat( [ reshape(f, shape)[pts,cells][:] for f in fields(var)]... )
  data = hcat(data_x, data_f)
  header = [ :x, [ fn for fn in fieldnames(typeof(var)) ]... ]
  n = pts.stop - pts.start + 1
  highlighters = []
  formatters = ft_printf("%0.12e")
  for j_cell in eachindex(cells)
    fnctr = (data, i, j) -> (i >= (j_cell-1)*n+1) && (i <= j_cell * n)
    clr = Crayon(foreground=cell_colors[periodic_index(j_cell, length(cell_colors))])
    push!(highlighters, Highlighter(fnctr, clr))
  end
  n_rows = size(data)[1]
  row_names = [ "" for i in 1:n_rows  ]
  for (i,cell) in enumerate(cells)
    row_names[(i-1)*n+1] = "$cell"
  end
  pretty_table(data, 
               header=header, 
               highlighters=Tuple(highlighters),
               formatters=formatters, 
               # body_hlines=collect(n:n:n_rows-n),
               crop=:none,
               row_names=row_names,
               row_name_column_title="#")
end

