# Design considerations
#
# - Settle on output directory structure
#
#   Example directory structure
#   advection_eq
#     script.jl ... run to generate data
#     sim_1
#       pars.toml
#       data.jld2
#       data.jld2#1
#       data.jld2#2
#       ....
#     sim_2
#       pars.toml
#       data.jld2
#       data.jld2#1
#       data.jld2#2
#       ....
#
#   #N tag for .jld2 files is inspired by DrWatson. 
#   I don't want to save parameters together with raw data, since this yields 
#   to duplication. The pars.toml file is saved so that one can rerun the
#   project and produce the same results, ideally (can this be really ensured
#   by Manifest.toml?)
#
#   - Q: Is this even a reasonable layout?
#     Q: What if parameter interfaces change? Should I start a new project
#     structure?
#   
# - Figure out where to put git commit hash and how to version outputs.
#   Possibilities
#     - Append git hash instead of #N tag to jld2 files.
#         +: Does not require to load file.
#         -: Unwieldly file names, not human readable, how to figure
#         out which git hash is the most recent? The most recent will be the
#         one that does not have a # tag. But still it does not say in which
#         order the other versions were created ...
#     - Add git hash as member to jld2 data
#         +: Can keep #N tag which is nicely human readable.
#         -: To retrieve git hash I have to (partly) load the jl2d file.
#         Q: What does 'partly load' mean? Doesn't JLD2 allow to access file
#         contents separately?
#         -: For DRY reasons, this would require to put all outputs into a 
#         single JLD2 file...
#     - Add git hash to .toml parameter file and also versionize those files
#         +: It is clear which file is the most recent and the git commit
#         will be human readable.
#         -: Again, duplication of parameters, since one sim folder is
#         supposed to only contain runs for a given configuration, but different
#         versions, there should be only one parameter file.
#           - Q: What to do when parameter interfaces change? Wouldn't this 
#           require to always maintain backwards compatibility for reproduceability
#           reasons? Or can I rely on git checkout?
#     - Somehow merge with log file.
#         +: This is where it actually belongs to. Also human readable.
#         -: Another file to generate (although I already planed to do it).
#         Log file has to be parsed in order to verify git hash which makes it 
#         more cumbersome to check if we are on the right commit. Although,
#         this could be easily done using regexs ...
#         Or maybe writing the log file in TOML language?
#     - Collect git hash together with version number labeling in a single
#       version.csv file that sits at the root of this directory.
#         +: Provided that results are only commited once for all
#         sims, then it provides a single place to look up the versions.
#         CSV works nice with DataFrames which I want to also use for data
#         analysis and to query and filter results.
#         -: Yet another file format besides .jld2, .toml, .txt (log file).
#
# - Q
#   - How often do I intend to run this pipeline?
#     Every few days, weeks or months.
#   - How often do I have to verify the git hash?
#     Whenever I want to analyze data from old runs.
#     - Is this a common occation?
#       Probably not, because who looks back?
#       Might only happen if I realize a bug and have to reanalyze the data.
#       Or in case I rerun an old version and want to check reproducability.
#   - Do I want to raise warnings or errors when loading versioned output?
#     Certainly, no.
#
# - Log file
#   - Contents to log
#       - Time step, ....
#     Actually, I don't have yet much to log for my simulations. 
#     Perhaps this changes in the future.
#
#   - Format
#     Log file should be human readable and since I already use yaml for the
#     parameter files, why not use it here in conjunction to store also
#     the git hash, method name, etc?
#     Thinking about it makes it clear that this is not the right solution,
#     since a log file does not need to be loaded ever again.
#     Better stick with plain txt file.
#
# - Workflow
#   - Given a clean working tree, modify script.jl to your needs (either add
#     more runs, or leave out certain configurations). Run simulations.
#     This should generate for each configuration processed log and output files
#     in the corresponding sim directories which are not yet tagged, but
#     come with a #WIP tag. 
#   - After making sure all results ran appropriately,
#     run a commit method that adds a new entry to the version.csv table with
#     the current tag number #N and git hash and also change the tags for
#     all #WIP files to #N.
#   - Commit changes to version control.


function get_version_table(outputdir)
  if !isdir(outputdir)
    error("'$outputdir' is not a directory!")
  end
  filename = joinpath(outputdir, "versions.csv")
  df = nothing
  if !isfile(filename)
    @warn "New version table found, generating new one ..."
    df = DataFrame(Id = Int[], GitHash = String[])
  else
    csv = CSV.File(filename, delim="")
    df = DataFrame(csv)
    sort!(df, [:Id])
  end
  return df
end


@doc """
Query git hash and check if repo is dirty.
Return nothing if hash retrival fails.
Inspired by DrWatson.jl.
"""
function githash(gitpath)
  repo = nothing
  try
    repo = LibGit2.GitRepoExt(gitpath)
  catch er
    if isa(er,LibGit2.GitError) && er.code == LibGit2.Error.ENOTFOUND
        @warn "The directory ('$gitpath') is not a Git repository. "*
        "returning `nothing` instead of the commit ID."
    else
        @warn "`gitdescribe` failed with '$er', returning `nothing` instead of the commit ID."
    end
    return nothing, false
  end
  return string(LibGit2.head_oid(repo)), LibGit2.isdirty(repo)
end


@doc """
Commit 
"""
function commit_simulations(outputdir)
  if !isdir(outputdir)
    error("'$outputdir' is not a directory!")
  end
  # retrieve git hash
  githash, isdirty = githash(outputdir)
  # check if new files were added
  datafiles = Glob.glob(r"^sim_\d+/data\.jld2\.#WIP$", direcotry=outputdir)
  if length(datafiles) == 0
    @warn "The output directory '$outputdir' does not contain WIP results."*
    "No entry committed."
    return
  end
  # check if output is complete (each directory should contain a log file, and
  # a jld2 file)
  wipfiles = []
  for datafile in datafiles
    logfile = joinpath(dirname(datafile), "logfile.txt.#WIP")
    if !isfile(logfile)
      @warn "Mismatch between number of log and data files encountered."*
      "No entry committed."
      return
    else
      push!(wipfiles, (datafile, logfile))
    end
  end
  # add new entry to dataframe
  nrows, _ = size(df)
  new_id = nrows + 1
  push!(df, [new_id, githash])
  # update tags from WIP to latest version
  for wipf in wipfiles
    for wipfile in wipf
      filename, _ = splitext(wipfile)
      cp(file, joinpath(filename, "#$new_id"))
    end
  end
  return
end
