@doc """
  `bdry_interpolating_polynomial_basis(xbase, w)`

Compute `w`th order polynomial basis `q_n` from Bugner 2017, eq. (3.78) evaluated on `xbase` grid.
Returned matrix contains nodal coefficients of `q_n` in rows.
`xbdrys` contains the subcell boundary coordinates.
"""
function bdry_interpolating_polynomial_basis(xbase, xbdrys)
  npts_bdry = length(xbdrys)
  n_bdry = npts_bdry - 1
  D = first_derivative_matrix(xbdrys)
  I = lagrange_interp_matrix(xbase, xbdrys)
  ID = I * D
  qs = zeros(n_bdry, length(xbase))
  for n = 0:n_bdry-1
    for m = n:n_bdry-1
      qs[n+1,:] .+= ID[:,m+2]
    end
  end
  qs .*= 2
  return qs
end



@doc """
  `optimal_weights(x, w)`

Compute optimal WENO weights matrix for a `2w+1` order reconstruction on a grid `x`,
where the weights for different cells are organized row-wise.

The computation follows Bugner 2017, section 3.3.1.
Note that the computation is done numerically in such a way, that all intermediate computations
are carried out including exact polynomial approxiamations.
"""
function optimal_weights(x, w)
  # w+1 subcells which all overlap with x
  xsubcells = [ [ 1.0 + 2.0 * (- w + l + k) for l = -1:w ] for k = 0:w ]
  # grid that spans full reconstruction stencile
  xfullcell = [ 1.0 + 2.0 * (- w + l ) for l = -1:2w ]

  ps_basis = [ bdry_interpolating_polynomial_basis(x, xcell) for xcell in xsubcells ]
  q_basis = bdry_interpolating_polynomial_basis(x, xfullcell)

  # for every point in x solve the least squares problem (3.81) from Bugner, p 43.
  npts = length(x)
  weights = zeros(w+1, npts)
  for idx_pt in 1:npts
    # build coefficient matrix from sub-cell-spanning polynomial basis
    A = zeros(2*w+1, w+1)
    for i_cell in 1:w+1
      A[i_cell:i_cell+w,i_cell] .= ps_basis[i_cell][:, idx_pt]
    end
    # build inhomogenity from full-cell-spanning polynomial basis
    b = q_basis[:, idx_pt]
    # solve least-squares problem for weights
    weights[:, idx_pt] .= A \ b
  end

  return weights
end



@doc """
  `smoothness_estimator_matrix(w)`

Compute matrix representation of smoothness estimator for an order `2w+1` WENO subcell splitting.
If `M` is this matrix, then smoothness estimation can be done by computing `u' * M * u`, where
`u` contain the `2w+1` sub cell averages.

Based on Bugner 2017, section 3.3.1.
Note that the computation is done numerically in such a way, that all intermediate computations
are carried out including exact polynomial approxiamations.
"""
function smoothness_estimator_matrix(w)

  @assert w <= 2 "Smoothness estimator only implemented up to w = 2"

  # setup a local quadrature rule to evaluate integrals over derivatives of lagrange polynomials
  element = SpectralElement(2*w, :LGL)
  @unpack iproduct, z, Npts = element

  # build w+1 subcell grids which all overlap with x
  xsubcells = [ [ 1.0 + 2.0 * (- w + l + k) for l = -1:w ] for k = 0:w ]
  # compute nodal basis coefficients for each subcell polynomial `p_i(x)`
  # Note that here `p_i` is represented as an order `w+1` polynomial, although it is actually
  # of order `w`m because it is the result of differentiating a `w+1` polynomial. We do so
  # in order to reuse the nodal derivative matrices.
  ps_basis = [ bdry_interpolating_polynomial_basis(xcell, xcell) for xcell in xsubcells ]

  # carry out sum of integrals over squared derivatives
  Es = [ zeros(w+1, w+1) for _ in 1:w+1 ] # smoothness estimator matrices for each subcell
  for (n, (xs, ps)) in enumerate(zip(xsubcells, ps_basis))

    # interpolate from subcell grid to quadrature grid
    I = dg1d.lagrange_interp_matrix(z, xs)

    # loop over derivatives
    for k in 1:w

      Dk = dg1d.derivative_matrix(k, xs)
      Lk = I * Dk # kth derivative of lagrange polynomial interpolated to quadrature grid
      integral_L2 = zeros(w+2,w+2) # because the ps' are represented as `w+1`th polynomials

      for i = 1:w+2, j = i:w+2 # collapsed loop because integral_L2 matrix is symmetric
        integral_L2[i,j] = iproduct(Lk[:,i], Lk[:,j])
        (i == j) && continue
        integral_L2[j,i] = integral_L2[i,j]
      end

      Es[n] .+= 2^(2*k-1) * ps * integral_L2 * transpose(ps)
    end

  end

  return Es
end



@with_kw struct Parameters_WENO <: AbstractParameters

  "Scheme order equals number of cells the reconstruction stencile covers around the troubled cell"
  order::Int64 = 3
  @assert order > 0 && isodd(order) "Only WENO schemes of odd and positive order are available"
  
  "Minimum value for inverse weighting of coefficients"
  threshold::Float64 = 1e-6
  @assert threshold > 0 "Threshold value must be positive"
end



struct WENOReconstruction
  prms::Parameters_WENO
  mesh::Mesh
  w::Int64
  subcell_bdry_coordinates::Vector{Vector{Float64}} # computational grid ⊂ [-2w+1,2w+1]
  subcell_polynomial_basis::Vector{Matrix{Float64}} # sampled on bdry coordinates
  optimal_weights::Matrix{Float64}
  smoothness_estimator::Vector{Matrix{Float64}}

  function WENOReconstruction(prms, mesh) 

    @unpack order = prms
    @unpack z = mesh.element

    w = Int64((order-1)/2)

    # build w+1 subcell grids which all overlap with x
    subcell_bdry_coordinates = [ [ 1.0 + 2.0 * (- w + l + k) for l = -1:w ] for k = 0:w ]
    subcell_polynomial_basis = [ bdry_interpolating_polynomial_basis(z, xcell) 
                                 for xcell in subcell_bdry_coordinates ]

    opt_weights = optimal_weights(z, w)
    smoothness_estimator = smoothness_estimator_matrix(w)

    return new(prms, mesh, w, subcell_bdry_coordinates, subcell_polynomial_basis,
              opt_weights, smoothness_estimator)
  end

end



@doc """
  `reconstruct!(u!::AbstractArray, trbld_idx_list::Vector{Int64}, weno::WENOReconstruction)`

Perform WENO reconstruction inplace on troubled subcells indexed by `trbld_idx_list`.
"""
function reconstruct!(u!::AbstractArray, trbld_idx_list::Vector{Int64}, weno::WENOReconstruction)

  @unpack w = weno
  @unpack K = weno.mesh
  @unpack Npts = weno.mesh.element
  shape = (Npts, K)

  mat_u! = reshape(u!, shape)
  mat_utmp = deepcopy(mat_u!)

  # cell-averages
  uavg = cellwise_average(mat_u!, weno.mesh)
  weights = zero(weno.optimal_weights)
  tmp_uweno = zeros(weno.mesh.element.Npts)

  # go through troubled cells
  for idx in trbld_idx_list

    @assert idx - w >= 1 && idx + w <= K "WENO reconstruction extends across bdry " *
      "which is not supported"

    uavg_stencile = uavg[idx-w:idx+w] # full width stencile surrounding troubled cell

    for idx = 1:w+1 # for every subcell
      subidx = 1+(idx-1):1+(idx-1)+w
      uavg_substencile = uavg_stencile[subidx] # covers ith reconstruction element

      # estimate smoothness, (3.69)
      smoothness = transpose(uavg_substencile) * weno.smoothness_estimator[idx] * uavg_substencile

      # compute weights, (3.70)
      @. weights[idx,:] = weno.optimal_weights[idx,:] / (weno.prms.threshold + smoothness)^2
    end

    # normalize weights, (3.71)
    weights ./= sum(weights, dims=1)
     
    # reconstruct solution, (3.72)
    fill!(tmp_uweno, 0.0)
    for idx = 1:w+1
      subidx = 1+(idx-1):1+(idx-1)+w
      uavg_substencile = uavg_stencile[subidx] # covers ith reconstruction element
      approx_poly = transpose(uavg_substencile) * weno.subcell_polynomial_basis[idx]
      @. tmp_uweno += weights[idx,:] * approx_poly[:]
    end

    # reinsert result
    mat_utmp[:, idx] = tmp_uweno
  end

  mat_u! .= mat_utmp

  return
end
