@testset "Utilities" begin


@testset "Least squares fit" begin

  npts = 57
  a = 5.7
  b = -12.4
  x = collect(Float64, 1:npts)
  y = @. a + b * x

  f_fit, a_fit, b_fit = dg1d.least_sqr_linear(x, y)

  @test a_fit ≈ a
  @test b_fit ≈ b

  # test with random data point
  i = Int(rand(x))
  @test f_fit(x[i]) ≈ y[i]

end


@testset "Newton Solver" begin

  f(x) = x^2 - π
  f_df(x) = tuple(x^2 - π, 2 * x)
  x0 = 1.1
  x, err, converged = dg1d.nlsolve(f, x0)
  @test converged == true
  @test x ≈ sqrt(π)
  x, err, converged = dg1d.nlsolve_exact(f_df, x0)
  @test converged == true
  @test x ≈ sqrt(π)

  # bad guess
  x0 = 0.0
  x, err, converged = dg1d.nlsolve(f, x0)
  @test converged == false

  # taken from wikipedia
  g(x) = cos(x) - x^3
  g_dg(x) = (cos(x) - x^3, - sin(x) - 3 * x^2)
  x0 = 0.5
  x, err, converged = dg1d.nlsolve(g, x0)
  @test converged = true
  x, err, converged = dg1d.nlsolve_exact(g_dg, x0)
  @test converged == true

end


@testset "FD differences" begin

  N = 2000
  L = 2.0
  dx = L / (N-1) # full step width

  # generate grid with randomly sampled grid spacings within range [0, L]
  x = [abs(randn())*dx]
  while true
    rnd = abs(randn())
    xi = dx * rnd + x[end]
    if xi > L
      break
    end
    push!(x, xi)
  end

  # test function
  f = @. x^2 - 2.0 * exp(-x^2)
  fx = @. 2.0 * x + 4.0 * exp(-x^2) * x
  fxx = @. 2.0 - 8.0 * exp(-x^2) * x^2 + 4.0 * exp(-x^2)

  nfx, nfxx = dg1d.fd_approx(f, x)
  xx = x[2:end-1]
  fx = fx[2:end-1]
  fxx = fxx[2:end-1]

  # think about reasonable numerical error to test against here
  # @test fx[2:end-1] ≈ nfx
  # @test fxx[2:end-1] ≈ nfxx

  # Plots.gr()
  # plt = Plots.plot()
  # Plots.plot!(plt, xx, fx - nfx, label="f_x (ana vs num)")
  # Plots.plot!(plt, xx, fxx - nfxx, label="f_xx (ana vs num)")

  # output_dir = dg1d.get_output_dir()
  # filename = "test_fd_approximation.pdf"
  # Plots.savefig(plt, joinpath(output_dir, filename))
  # @info "Test figure saved to $filename ..."

end


end
